@@ -8,10 +8,16 @@ ARCH=$2
 TMPDIR=$3
 
 if [ "$GOVERSION" = "gotip" ]; then
-    exit 0
-    bootstrapver=$(curl https://golang.org/VERSION?m=text)
+    bootstrapver=$(curl https://go.dev/VERSION?m=text)
+    cd $TMPDIR
     curl -sSL "https://storage.googleapis.com/golang/$bootstrapver.darwin-$ARCH.tar.gz" | tar -xz
-    git clone https://go.googlesource.com/go $TMPDIR/go-tip
+    cd -
+    if [ -x $TMPDIR/go-tip ]; then
+    	cd $TMPDIR/go-tip
+    	git pull origin
+    else
+    	git clone https://go.googlesource.com/go $TMPDIR/go-tip
+    fi
     export GOROOT_BOOTSTRAP=$TMPDIR/go
     export GOROOT=$TMPDIR/go-tip
     cd $TMPDIR/go-tip/src
@@ -29,6 +35,8 @@ fi
 
 mkdir -p $TMPDIR/gopath
 
+go env
+
 export GOPATH="$TMPDIR/gopath"
 export GOARCH="$ARCH"
 export PATH="$GOROOT/bin:$PATH"
