@@ -164,3 +164,20 @@ func (reg *DwarfRegister) FillBytes() {
 	reg.Bytes = make([]byte, 8)
 	binary.LittleEndian.PutUint64(reg.Bytes, reg.Uint64Val)
 }
+
+// Overwrite takes the contents of reg1 and overwrites them with the contents
+// of reg2 in little-endian order, returning a new register. The new register
+// will always contain the complete contents of both registers, so if reg2 is
+// larger than reg1, the final register will be reg2's size.
+func (reg1 *DwarfRegister) Overwrite(reg2 *DwarfRegister) *DwarfRegister {
+	reg1.FillBytes()
+	reg2.FillBytes()
+	width := len(reg1.Bytes)
+	if len(reg2.Bytes) > len(reg1.Bytes) {
+		width = len(reg2.Bytes)
+	}
+	b := make([]byte, width)
+	copy(b, reg1.Bytes)
+	copy(b, reg2.Bytes)
+	return DwarfRegisterFromBytes(b)
+}
