@@ -30,6 +30,7 @@ import (
 	"github.com/go-delve/delve/pkg/dwarf/util"
 	"github.com/go-delve/delve/pkg/goversion"
 	"github.com/go-delve/delve/pkg/logflags"
+	"github.com/go-delve/delve/pkg/proc/debuginfod"
 	"github.com/hashicorp/golang-lru/simplelru"
 	"github.com/sirupsen/logrus"
 )
@@ -51,12 +52,18 @@ type BinaryInfo struct {
 
 	debugInfoDirectories []string
 
+	// BuildID of this binary.
+	BuildID string
+
 	// Functions is a list of all DW_TAG_subprogram entries in debug_info, sorted by entry point
 	Functions []Function
 	// Sources is a list of all source files found in debug_line.
 	Sources []string
 	// LookupFunc maps function names to a description of the function.
 	LookupFunc map[string]*Function
+	// lookupGenericFunc maps function names, with their type parameters removed, to functions.
+	// Functions that are not generic are not added to this map.
+	lookupGenericFunc map[string][]*Function
 
 	// SymNames maps addr to a description *elf.Symbol of this addr.
 	SymNames map[uint64]*elf.Symbol
@@ -100,6 +107,12 @@ type BinaryInfo struct {
 	// function starts.
 	inlinedCallLines map[fileLine][]uint64
 
+	// dwrapUnwrapCache caches unwrapping of defer wrapper functions (dwrap)
+	dwrapUnwrapCache map[uint64]*Function
+
+	// Go 1.17 register ABI is enabled.
+	regabi bool
+
 	logger *logrus.Entry
 }
 
@@ -142,34 +155,192 @@ func (err *ErrFunctionNotFound) Error() string {
 
 // FindFileLocation returns the PC for a given file:line.
 // Assumes that `file` is normalized to lower case and '/' on Windows.
-func FindFileLocation(p Process, fileName string, lineno int) ([]uint64, error) {
-	pcs, err := p.BinInfo().LineToPC(fileName, lineno)
-	if err != nil {
-		return nil, err
+func FindFileLocation(p Process, filename string, lineno int) ([]uint64, error) {
+	// A single file:line can appear in multiple concrete functions, because of
+	// generics instantiation as well as multiple inlined calls into other
+	// concrete functions.
+
+	// 1. Find all instructions assigned in debug_line to filename:lineno.
+
+	bi := p.BinInfo()
+
+	fileFound := false
+	pcs := []line.PCStmt{}
+	for _, image := range bi.Images {
+		for _, cu := range image.compileUnits {
+			if cu.lineInfo == nil || cu.lineInfo.Lookup[filename] == nil {
+				continue
+			}
+
+			fileFound = true
+			pcs = append(pcs, cu.lineInfo.LineToPCs(filename, lineno)...)
+		}
 	}
+
+	if len(pcs) == 0 {
+		// Check if the line contained a call to a function that was inlined, in
+		// that case it's possible for the line itself to not appear in debug_line
+		// at all, but it will still be in debug_info as the call site for an
+		// inlined subroutine entry.
+		for _, pc := range bi.inlinedCallLines[fileLine{filename, lineno}] {
+			pcs = append(pcs, line.PCStmt{PC: pc, Stmt: true})
+		}
+	}
+
+	if len(pcs) == 0 {
+		return nil, &ErrCouldNotFindLine{fileFound, filename, lineno}
+	}
+
+	// 2. assign all occurences of filename:lineno to their containing function
+
+	pcByFunc := map[*Function][]line.PCStmt{}
+	sort.Slice(pcs, func(i, j int) bool { return pcs[i].PC < pcs[j].PC })
 	var fn *Function
-	for i := range pcs {
-		if fn == nil || pcs[i] < fn.Entry || pcs[i] >= fn.End {
-			fn = p.BinInfo().PCToFunc(pcs[i])
+	for _, pcstmt := range pcs {
+		if fn == nil || (pcstmt.PC < fn.Entry) || (pcstmt.PC >= fn.End) {
+			fn = p.BinInfo().PCToFunc(pcstmt.PC)
 		}
-		if fn != nil && fn.Entry == pcs[i] {
-			pcs[i], _ = FirstPCAfterPrologue(p, fn, true)
+		if fn != nil {
+			pcByFunc[fn] = append(pcByFunc[fn], pcstmt)
 		}
 	}
-	return pcs, nil
+
+	selectedPCs := []uint64{}
+
+	for fn, pcs := range pcByFunc {
+
+		// 3. for each concrete function split instruction between the inlined functions it contains
+
+		if strings.Contains(fn.Name, "·dwrap·") || fn.trampoline {
+			// skip autogenerated functions
+			continue
+		}
+
+		dwtree, err := fn.cu.image.getDwarfTree(fn.offset)
+		if err != nil {
+			return nil, fmt.Errorf("loading DWARF for %s@%#x: %v", fn.Name, fn.offset, err)
+		}
+		inlrngs := allInlineCallRanges(dwtree)
+
+		// findInlRng returns the DWARF offset of the inlined call containing pc.
+		// If multiple nested inlined calls contain pc the deepest one is returned
+		// (since allInlineCallRanges returns inlined call by decreasing depth
+		// this is the first matching entry of the slice).
+		findInlRng := func(pc uint64) dwarf.Offset {
+			for _, inlrng := range inlrngs {
+				if inlrng.rng[0] <= pc && pc < inlrng.rng[1] {
+					return inlrng.off
+				}
+			}
+			return fn.offset
+		}
+
+		pcsByOff := map[dwarf.Offset][]line.PCStmt{}
+
+		for _, pc := range pcs {
+			off := findInlRng(pc.PC)
+			pcsByOff[off] = append(pcsByOff[off], pc)
+		}
+
+		// 4. pick the first instruction with stmt set for each inlined call as
+		//    well as the main body of the concrete function. If nothing has
+		//    is_stmt set pick the first instruction instead.
+
+		for off, pcs := range pcsByOff {
+			sort.Slice(pcs, func(i, j int) bool { return pcs[i].PC < pcs[j].PC })
+
+			var selectedPC uint64
+			for _, pc := range pcs {
+				if pc.Stmt {
+					selectedPC = pc.PC
+					break
+				}
+			}
+
+			if selectedPC == 0 && len(pcs) > 0 {
+				selectedPC = pcs[0].PC
+			}
+
+			if selectedPC == 0 {
+				continue
+			}
+
+			// 5. if we picked the entry point of the function, skip it
+
+			if off == fn.offset && fn.Entry == selectedPC {
+				selectedPC, _ = FirstPCAfterPrologue(p, fn, true)
+			}
+
+			selectedPCs = append(selectedPCs, selectedPC)
+		}
+	}
+
+	sort.Slice(selectedPCs, func(i, j int) bool { return selectedPCs[i] < selectedPCs[j] })
+
+	return selectedPCs, nil
+}
+
+// inlRnage is the range of an inlined call
+type inlRange struct {
+	off   dwarf.Offset
+	depth uint32
+	rng   [2]uint64
+}
+
+// allInlineCallRanges returns all inlined calls contained inside 'tree' in
+// reverse nesting order (i.e. the most nested calls are returned first).
+// Note that a single inlined call might not have a continuous range of
+// addresses and therefore appear multiple times in the returned slice.
+func allInlineCallRanges(tree *godwarf.Tree) []inlRange {
+	r := []inlRange{}
+
+	var visit func(*godwarf.Tree, uint32)
+	visit = func(n *godwarf.Tree, depth uint32) {
+		if n.Tag == dwarf.TagInlinedSubroutine {
+			for _, rng := range n.Ranges {
+				r = append(r, inlRange{off: n.Offset, depth: depth, rng: rng})
+			}
+		}
+		for _, child := range n.Children {
+			visit(child, depth+1)
+		}
+	}
+	visit(tree, 0)
+
+	sort.SliceStable(r, func(i, j int) bool { return r[i].depth > r[j].depth })
+	return r
+}
+
+// FindFunction returns the functions with name funcName.
+func (bi *BinaryInfo) FindFunction(funcName string) ([]*Function, error) {
+	if fn := bi.LookupFunc[funcName]; fn != nil {
+		return []*Function{fn}, nil
+	}
+	fns := bi.LookupGenericFunc()[funcName]
+	if len(fns) == 0 {
+		return nil, &ErrFunctionNotFound{funcName}
+	}
+	return fns, nil
 }
 
 // FindFunctionLocation finds address of a function's line
 // If lineOffset is passed FindFunctionLocation will return the address of that line
 func FindFunctionLocation(p Process, funcName string, lineOffset int) ([]uint64, error) {
 	bi := p.BinInfo()
-	origfn := bi.LookupFunc[funcName]
-	if origfn == nil {
-		return nil, &ErrFunctionNotFound{funcName}
+	origfns, err := bi.FindFunction(funcName)
+	if err != nil {
+		return nil, err
 	}
 
-	if lineOffset <= 0 {
-		r := make([]uint64, 0, len(origfn.InlinedCalls)+1)
+	if lineOffset > 0 {
+		fn := origfns[0]
+		filename, lineno := fn.cu.lineInfo.PCToLine(fn.Entry, fn.Entry)
+		return FindFileLocation(p, filename, lineno+lineOffset)
+	}
+
+	r := make([]uint64, 0, len(origfns[0].InlinedCalls)+len(origfns))
+
+	for _, origfn := range origfns {
 		if origfn.Entry > 0 {
 			// add concrete implementation of the function
 			pc, err := FirstPCAfterPrologue(p, origfn, false)
@@ -185,10 +356,9 @@ func FindFunctionLocation(p Process, funcName string, lineOffset int) ([]uint64,
 		if len(r) == 0 {
 			return nil, &ErrFunctionNotFound{funcName}
 		}
-		return r, nil
 	}
-	filename, lineno := origfn.cu.lineInfo.PCToLine(origfn.Entry, origfn.Entry)
-	return bi.LineToPC(filename, lineno+lineOffset)
+	sort.Slice(r, func(i, j int) bool { return r[i] < r[j] })
+	return r, nil
 }
 
 // FirstPCAfterPrologue returns the address of the first
@@ -239,15 +409,15 @@ func (e *ErrUnsupportedArch) Error() string {
 	var supportArchs []cpuArch
 	switch e.os {
 	case "linux":
-		for linuxArch, _ := range supportedLinuxArch {
+		for linuxArch := range supportedLinuxArch {
 			supportArchs = append(supportArchs, linuxArch)
 		}
 	case "windows":
-		for windowArch, _ := range supportedWindowsArch {
+		for windowArch := range supportedWindowsArch {
 			supportArchs = append(supportArchs, windowArch)
 		}
 	case "darwin":
-		for darwinArch, _ := range supportedDarwinArch {
+		for darwinArch := range supportedDarwinArch {
 			supportArchs = append(supportArchs, darwinArch)
 		}
 	}
@@ -307,15 +477,41 @@ type Function struct {
 	offset     dwarf.Offset
 	cu         *compileUnit
 
+	trampoline bool // DW_AT_trampoline attribute set to true
+
 	// InlinedCalls lists all inlined calls to this function
 	InlinedCalls []InlinedCall
 }
 
+// instRange returns the indexes in fn.Name of the type parameter
+// instantiation, which is the position of the outermost '[' and ']'.
+// If fn is not an instantiated function both returned values will be len(fn.Name)
+func (fn *Function) instRange() [2]int {
+	d := len(fn.Name)
+	inst := [2]int{d, d}
+	if strings.HasPrefix(fn.Name, "type..") {
+		return inst
+	}
+	inst[0] = strings.Index(fn.Name, "[")
+	if inst[0] < 0 {
+		inst[0] = d
+		return inst
+	}
+	inst[1] = strings.LastIndex(fn.Name, "]")
+	if inst[1] < 0 {
+		inst[0] = d
+		inst[1] = d
+		return inst
+	}
+	return inst
+}
+
 // PackageName returns the package part of the symbol name,
 // or the empty string if there is none.
 // Borrowed from $GOROOT/debug/gosym/symtab.go
 func (fn *Function) PackageName() string {
-	return packageName(fn.Name)
+	inst := fn.instRange()
+	return packageName(fn.Name[:inst[0]])
 }
 
 func packageName(name string) string {
@@ -334,27 +530,44 @@ func packageName(name string) string {
 // or the empty string if there is none.
 // Borrowed from $GOROOT/debug/gosym/symtab.go
 func (fn *Function) ReceiverName() string {
-	pathend := strings.LastIndex(fn.Name, "/")
+	inst := fn.instRange()
+	pathend := strings.LastIndex(fn.Name[:inst[0]], "/")
 	if pathend < 0 {
 		pathend = 0
 	}
 	l := strings.Index(fn.Name[pathend:], ".")
-	r := strings.LastIndex(fn.Name[pathend:], ".")
-	if l == -1 || r == -1 || l == r {
+	if l == -1 {
 		return ""
 	}
-	return fn.Name[pathend+l+1 : pathend+r]
+	if r := strings.LastIndex(fn.Name[inst[1]:], "."); r != -1 && pathend+l != inst[1]+r {
+		return fn.Name[pathend+l+1 : inst[1]+r]
+	} else if r := strings.LastIndex(fn.Name[pathend:inst[0]], "."); r != -1 && l != r {
+		return fn.Name[pathend+l+1 : pathend+r]
+	}
+	return ""
 }
 
 // BaseName returns the symbol name without the package or receiver name.
 // Borrowed from $GOROOT/debug/gosym/symtab.go
 func (fn *Function) BaseName() string {
-	if i := strings.LastIndex(fn.Name, "."); i != -1 {
+	inst := fn.instRange()
+	if i := strings.LastIndex(fn.Name[inst[1]:], "."); i != -1 {
+		return fn.Name[inst[1]+i+1:]
+	} else if i := strings.LastIndex(fn.Name[:inst[0]], "."); i != -1 {
 		return fn.Name[i+1:]
 	}
 	return fn.Name
 }
 
+// NameWithoutTypeParams returns the function name without instantiation parameters
+func (fn *Function) NameWithoutTypeParams() string {
+	inst := fn.instRange()
+	if inst[0] == inst[1] {
+		return fn.Name
+	}
+	return fn.Name[:inst[0]] + fn.Name[inst[1]+1:]
+}
+
 // Optimized returns true if the function was optimized by the compiler.
 func (fn *Function) Optimized() bool {
 	return fn.cu.optimized
@@ -512,69 +725,6 @@ func (err *ErrCouldNotFindLine) Error() string {
 	return fmt.Sprintf("could not find file %s", err.filename)
 }
 
-// LineToPC converts a file:line into a list of matching memory addresses,
-// corresponding to the first instruction matching the specified file:line
-// in the containing function and all its inlined calls.
-func (bi *BinaryInfo) LineToPC(filename string, lineno int) (pcs []uint64, err error) {
-	fileFound := false
-	var pc uint64
-pcsearch:
-	for _, image := range bi.Images {
-		for _, cu := range image.compileUnits {
-			if cu.lineInfo == nil || cu.lineInfo.Lookup[filename] == nil {
-				continue
-			}
-			fileFound = true
-			pc = cu.lineInfo.LineToPC(filename, lineno)
-			if pc != 0 {
-				break pcsearch
-			}
-		}
-	}
-
-	if pc == 0 {
-		// Check if the line contained a call to a function that was inlined, in
-		// that case it's possible for the line itself to not appear in debug_line
-		// at all, but it will still be in debug_info as the call site for an
-		// inlined subroutine entry.
-		if pcs := bi.inlinedCallLines[fileLine{filename, lineno}]; len(pcs) != 0 {
-			return pcs, nil
-		}
-		return nil, &ErrCouldNotFindLine{fileFound, filename, lineno}
-	}
-	// The code above will find the first occurence of an instruction
-	// corresponding to filename:line. If the function corresponding to that
-	// instruction has been inlined we don't just want to return the first
-	// occurence (which could be either the concrete version of the function or
-	// one of the inlinings) but instead:
-	// - the first instruction corresponding to filename:line in the concrete
-	//   version of the function
-	// - the first instruction corresponding to filename:line in each inlined
-	//   instance of the function.
-	fn := bi.PCToInlineFunc(pc)
-	if fn == nil {
-		return []uint64{pc}, nil
-	}
-	pcs = make([]uint64, 0, len(fn.InlinedCalls)+1)
-	pcs = appendLineToPCIn(pcs, filename, lineno, fn.cu, fn, fn.Entry, fn.End)
-	for _, call := range fn.InlinedCalls {
-		pcs = appendLineToPCIn(pcs, filename, lineno, call.cu, bi.PCToFunc(call.LowPC), call.LowPC, call.HighPC)
-	}
-	return pcs, nil
-}
-
-func appendLineToPCIn(pcs []uint64, filename string, lineno int, cu *compileUnit, containingFn *Function, lowPC, highPC uint64) []uint64 {
-	var entry uint64
-	if containingFn != nil {
-		entry = containingFn.Entry
-	}
-	pc := cu.lineInfo.LineToPCIn(filename, lineno, entry, lowPC, highPC)
-	if pc != 0 {
-		return append(pcs, pc)
-	}
-	return pcs
-}
-
 // AllPCsForFileLines returns a map providing all PC addresses for filename and each line in linenos
 func (bi *BinaryInfo) AllPCsForFileLines(filename string, linenos []int) map[int][]uint64 {
 	r := make(map[int][]uint64)
@@ -607,27 +757,6 @@ func (bi *BinaryInfo) PCToFunc(pc uint64) *Function {
 	return nil
 }
 
-// PCToInlineFunc returns the function containing the given PC address.
-// If the PC address belongs to an inlined call it will return the inlined function.
-func (bi *BinaryInfo) PCToInlineFunc(pc uint64) *Function {
-	fn := bi.PCToFunc(pc)
-	dwarfTree, err := fn.cu.image.getDwarfTree(fn.offset)
-	if err != nil {
-		return fn
-	}
-	entries := reader.InlineStack(dwarfTree, pc)
-	if len(entries) == 0 {
-		return fn
-	}
-
-	fnname, okname := entries[0].Val(dwarf.AttrName).(string)
-	if !okname {
-		return fn
-	}
-
-	return bi.LookupFunc[fnname]
-}
-
 // PCToImage returns the image containing the given PC address.
 func (bi *BinaryInfo) PCToImage(pc uint64) *Image {
 	fn := bi.PCToFunc(pc)
@@ -645,17 +774,19 @@ type Image struct {
 	closer         io.Closer
 	sepDebugCloser io.Closer
 
-	dwarf       *dwarf.Data
-	dwarfReader *dwarf.Reader
-	loclist2    *loclist.Dwarf2Reader
-	loclist5    *loclist.Dwarf5Reader
-	debugAddr   *godwarf.DebugAddrSection
+	dwarf        *dwarf.Data
+	dwarfReader  *dwarf.Reader
+	loclist2     *loclist.Dwarf2Reader
+	loclist5     *loclist.Dwarf5Reader
+	debugAddr    *godwarf.DebugAddrSection
+	debugLineStr []byte
 
 	typeCache map[dwarf.Offset]godwarf.Type
 
 	compileUnits []*compileUnit // compileUnits is sorted by increasing DWARF offset
 
-	dwarfTreeCache *simplelru.LRU
+	dwarfTreeCache      *simplelru.LRU
+	runtimeMallocgcTree *godwarf.Tree // patched version of runtime.mallocgc's DIE
 
 	// runtimeTypeToDIE maps between the offset of a runtime._type in
 	// runtime.moduledata.types and the offset of the DIE in debug_info. This
@@ -670,7 +801,7 @@ type Image struct {
 func (image *Image) registerRuntimeTypeToDIE(entry *dwarf.Entry, ardr *reader.Reader) {
 	if off, ok := entry.Val(godwarf.AttrGoRuntimeType).(uint64); ok {
 		if _, ok := image.runtimeTypeToDIE[off]; !ok {
-			image.runtimeTypeToDIE[off+image.StaticBase] = runtimeTypeDIE{entry.Offset, -1}
+			image.runtimeTypeToDIE[off] = runtimeTypeDIE{entry.Offset, -1}
 		}
 	}
 }
@@ -707,14 +838,28 @@ func (bi *BinaryInfo) AddImage(path string, addr uint64) error {
 
 // moduleDataToImage finds the image corresponding to the given module data object.
 func (bi *BinaryInfo) moduleDataToImage(md *moduleData) *Image {
-	return bi.funcToImage(bi.PCToFunc(uint64(md.text)))
+	fn := bi.PCToFunc(uint64(md.text))
+	if fn != nil {
+		return bi.funcToImage(fn)
+	}
+	// Try searching for the image with the closest address preceding md.text
+	var so *Image
+	for i := range bi.Images {
+		if int64(bi.Images[i].StaticBase) > int64(md.text) {
+			continue
+		}
+		if so == nil || int64(bi.Images[i].StaticBase) > int64(so.StaticBase) {
+			so = bi.Images[i]
+		}
+	}
+	return so
 }
 
 // imageToModuleData finds the module data in mds corresponding to the given image.
 func (bi *BinaryInfo) imageToModuleData(image *Image, mds []moduleData) *moduleData {
 	for _, md := range mds {
 		im2 := bi.moduleDataToImage(&md)
-		if im2.index == image.index {
+		if im2 != nil && im2.index == image.index {
 			return &md
 		}
 	}
@@ -769,10 +914,13 @@ func (image *Image) Close() error {
 	return err2
 }
 
-func (image *Image) setLoadError(fmtstr string, args ...interface{}) {
+func (image *Image) setLoadError(logger *logrus.Entry, fmtstr string, args ...interface{}) {
 	image.loadErrMu.Lock()
 	image.loadErr = fmt.Errorf(fmtstr, args...)
 	image.loadErrMu.Unlock()
+	if logger != nil {
+		logger.Errorf("error loading binary %q: %v", image.Path, image.loadErr)
+	}
 }
 
 // LoadError returns any error incurred while loading this image.
@@ -781,6 +929,9 @@ func (image *Image) LoadError() error {
 }
 
 func (image *Image) getDwarfTree(off dwarf.Offset) (*godwarf.Tree, error) {
+	if image.runtimeMallocgcTree != nil && off == image.runtimeMallocgcTree.Offset {
+		return image.runtimeMallocgcTree, nil
+	}
 	if r, ok := image.dwarfTreeCache.Get(off); ok {
 		return r.(*godwarf.Tree), nil
 	}
@@ -911,12 +1062,16 @@ func (bi *BinaryInfo) LocationCovers(entry *dwarf.Entry, attr dwarf.Attr) ([][2]
 // This will either be an int64 address or a slice of Pieces for locations
 // that don't correspond to a single memory address (registers, composite
 // locations).
-func (bi *BinaryInfo) Location(entry godwarf.Entry, attr dwarf.Attr, pc uint64, regs op.DwarfRegisters) (int64, []op.Piece, *locationExpr, error) {
+func (bi *BinaryInfo) Location(entry godwarf.Entry, attr dwarf.Attr, pc uint64, regs op.DwarfRegisters, mem MemoryReadWriter) (int64, []op.Piece, *locationExpr, error) {
 	instr, descr, err := bi.locationExpr(entry, attr, pc)
 	if err != nil {
 		return 0, nil, nil, err
 	}
-	addr, pieces, err := op.ExecuteStackProgram(regs, instr, bi.Arch.PtrSize())
+	readMemory := op.ReadMemoryFunc(nil)
+	if mem != nil {
+		readMemory = mem.ReadMemory
+	}
+	addr, pieces, err := op.ExecuteStackProgram(regs, instr, bi.Arch.PtrSize(), readMemory)
 	return addr, pieces, descr, err
 }
 
@@ -1012,14 +1167,14 @@ func (bi *BinaryInfo) funcToImage(fn *Function) *Image {
 // debug_frame is present it must be parsable correctly.
 func (bi *BinaryInfo) parseDebugFrameGeneral(image *Image, debugFrameBytes []byte, debugFrameName string, debugFrameErr error, ehFrameBytes []byte, ehFrameAddr uint64, ehFrameName string, byteOrder binary.ByteOrder) {
 	if debugFrameBytes == nil && ehFrameBytes == nil {
-		image.setLoadError("could not get %s section: %v", debugFrameName, debugFrameErr)
+		image.setLoadError(bi.logger, "could not get %s section: %v", debugFrameName, debugFrameErr)
 		return
 	}
 
 	if debugFrameBytes != nil {
 		fe, err := frame.Parse(debugFrameBytes, byteOrder, image.StaticBase, bi.Arch.PtrSize(), 0)
 		if err != nil {
-			image.setLoadError("could not parse %s section: %v", debugFrameName, err)
+			image.setLoadError(bi.logger, "could not parse %s section: %v", debugFrameName, err)
 			return
 		}
 		bi.frameEntries = bi.frameEntries.Append(fe)
@@ -1029,7 +1184,7 @@ func (bi *BinaryInfo) parseDebugFrameGeneral(image *Image, debugFrameBytes []byt
 		fe, err := frame.Parse(ehFrameBytes, byteOrder, image.StaticBase, bi.Arch.PtrSize(), ehFrameAddr)
 		if err != nil {
 			if debugFrameBytes == nil {
-				image.setLoadError("could not parse %s section: %v", ehFrameName, err)
+				image.setLoadError(bi.logger, "could not parse %s section: %v", ehFrameName, err)
 				return
 			}
 			bi.logger.Warnf("could not parse %s section: %v", ehFrameName, err)
@@ -1041,15 +1196,6 @@ func (bi *BinaryInfo) parseDebugFrameGeneral(image *Image, debugFrameBytes []byt
 
 // ELF ///////////////////////////////////////////////////////////////
 
-// ErrNoBuildIDNote is used in openSeparateDebugInfo to signal there's no
-// build-id note on the binary, so LoadBinaryInfoElf will return
-// the error message coming from elfFile.DWARF() instead.
-type ErrNoBuildIDNote struct{}
-
-func (e *ErrNoBuildIDNote) Error() string {
-	return "can't find build-id note on binary"
-}
-
 // openSeparateDebugInfo searches for a file containing the separate
 // debug info for the binary using the "build ID" method as described
 // in GDB's documentation [1], and if found returns two handles, one
@@ -1060,14 +1206,11 @@ func (e *ErrNoBuildIDNote) Error() string {
 // will look in directories specified by the debug-info-directories config value.
 func (bi *BinaryInfo) openSeparateDebugInfo(image *Image, exe *elf.File, debugInfoDirectories []string) (*os.File, *elf.File, error) {
 	var debugFilePath string
+	var err error
 	for _, dir := range debugInfoDirectories {
 		var potentialDebugFilePath string
 		if strings.Contains(dir, "build-id") {
-			desc1, desc2, err := parseBuildID(exe)
-			if err != nil {
-				continue
-			}
-			potentialDebugFilePath = fmt.Sprintf("%s/%s/%s.debug", dir, desc1, desc2)
+			potentialDebugFilePath = fmt.Sprintf("%s/%s/%s.debug", dir, bi.BuildID[:2], bi.BuildID[2:])
 		} else if strings.HasPrefix(image.Path, "/proc") {
 			path, err := filepath.EvalSymlinks(image.Path)
 			if err == nil {
@@ -1082,8 +1225,13 @@ func (bi *BinaryInfo) openSeparateDebugInfo(image *Image, exe *elf.File, debugIn
 			break
 		}
 	}
+	// We cannot find the debug information locally on the system. Try and see if we're on a system that
+	// has debuginfod so that we can use that in order to find any relevant debug information.
 	if debugFilePath == "" {
-		return nil, nil, ErrNoDebugInfoFound
+		debugFilePath, err = debuginfod.GetDebuginfo(bi.BuildID)
+		if err != nil {
+			return nil, nil, ErrNoDebugInfoFound
+		}
 	}
 	sepFile, err := os.OpenFile(debugFilePath, 0, os.ModePerm)
 	if err != nil {
@@ -1104,35 +1252,6 @@ func (bi *BinaryInfo) openSeparateDebugInfo(image *Image, exe *elf.File, debugIn
 	return sepFile, elfFile, nil
 }
 
-func parseBuildID(exe *elf.File) (string, string, error) {
-	buildid := exe.Section(".note.gnu.build-id")
-	if buildid == nil {
-		return "", "", &ErrNoBuildIDNote{}
-	}
-
-	br := buildid.Open()
-	bh := new(buildIDHeader)
-	if err := binary.Read(br, binary.LittleEndian, bh); err != nil {
-		return "", "", errors.New("can't read build-id header: " + err.Error())
-	}
-
-	name := make([]byte, bh.Namesz)
-	if err := binary.Read(br, binary.LittleEndian, name); err != nil {
-		return "", "", errors.New("can't read build-id name: " + err.Error())
-	}
-
-	if strings.TrimSpace(string(name)) != "GNU\x00" {
-		return "", "", errors.New("invalid build-id signature")
-	}
-
-	descBinary := make([]byte, bh.Descsz)
-	if err := binary.Read(br, binary.LittleEndian, descBinary); err != nil {
-		return "", "", errors.New("can't read build-id desc: " + err.Error())
-	}
-	desc := hex.EncodeToString(descBinary)
-	return desc[:2], desc[2:], nil
-}
-
 // loadBinaryInfoElf specifically loads information from an ELF binary.
 func loadBinaryInfoElf(bi *BinaryInfo, image *Image, path string, addr uint64, wg *sync.WaitGroup) error {
 	exe, err := os.OpenFile(path, 0, os.ModePerm)
@@ -1169,6 +1288,7 @@ func loadBinaryInfoElf(bi *BinaryInfo, image *Image, path string, addr uint64, w
 
 	dwarfFile := elfFile
 
+	bi.loadBuildID(image, elfFile)
 	var debugInfoBytes []byte
 	image.dwarf, err = elfFile.DWARF()
 	if err != nil {
@@ -1202,9 +1322,11 @@ func loadBinaryInfoElf(bi *BinaryInfo, image *Image, path string, addr uint64, w
 	image.loclist5 = loclist.NewDwarf5Reader(debugLoclistBytes)
 	debugAddrBytes, _ := godwarf.GetDebugSectionElf(dwarfFile, "addr")
 	image.debugAddr = godwarf.ParseAddr(debugAddrBytes)
+	debugLineStrBytes, _ := godwarf.GetDebugSectionElf(dwarfFile, "line_str")
+	image.debugLineStr = debugLineStrBytes
 
 	wg.Add(3)
-	go bi.parseDebugFrameElf(image, dwarfFile, debugInfoBytes, wg)
+	go bi.parseDebugFrameElf(image, dwarfFile, elfFile, debugInfoBytes, wg)
 	go bi.loadDebugInfoMaps(image, debugInfoBytes, debugLineBytes, wg, nil)
 	go bi.loadSymbolName(image, elfFile, wg)
 	if image.index == 0 {
@@ -1224,21 +1346,52 @@ func (bi *BinaryInfo) loadSymbolName(image *Image, file *elf.File, wg *sync.Wait
 		bi.SymNames = make(map[uint64]*elf.Symbol)
 	}
 	symSecs, _ := file.Symbols()
-	if symSecs != nil {
-		for _, symSec := range symSecs {
-			if symSec.Info == _STT_FUNC { // TODO(chainhelen), need to parse others types.
-				s := symSec
-				bi.SymNames[symSec.Value+image.StaticBase] = &s
-			}
+	for _, symSec := range symSecs {
+		if symSec.Info == _STT_FUNC { // TODO(chainhelen), need to parse others types.
+			s := symSec
+			bi.SymNames[symSec.Value+image.StaticBase] = &s
 		}
 	}
 }
 
-func (bi *BinaryInfo) parseDebugFrameElf(image *Image, exe *elf.File, debugInfoBytes []byte, wg *sync.WaitGroup) {
+func (bi *BinaryInfo) loadBuildID(image *Image, file *elf.File) {
+	buildid := file.Section(".note.gnu.build-id")
+	if buildid == nil {
+		bi.logger.Error("can't find build-id note on binary")
+		return
+	}
+
+	br := buildid.Open()
+	bh := new(buildIDHeader)
+	if err := binary.Read(br, binary.LittleEndian, bh); err != nil {
+		bi.logger.Errorf("can't read build-id header: %v", err)
+		return
+	}
+
+	name := make([]byte, bh.Namesz)
+	if err := binary.Read(br, binary.LittleEndian, name); err != nil {
+		bi.logger.Errorf("can't read build-id name: %v", err)
+		return
+	}
+
+	if strings.TrimSpace(string(name)) != "GNU\x00" {
+		bi.logger.Error("invalid build-id signature")
+		return
+	}
+
+	descBinary := make([]byte, bh.Descsz)
+	if err := binary.Read(br, binary.LittleEndian, descBinary); err != nil {
+		bi.logger.Errorf("can't read build-id desc: %v", err)
+		return
+	}
+	bi.BuildID = hex.EncodeToString(descBinary)
+}
+
+func (bi *BinaryInfo) parseDebugFrameElf(image *Image, dwarfFile, exeFile *elf.File, debugInfoBytes []byte, wg *sync.WaitGroup) {
 	defer wg.Done()
 
-	debugFrameData, debugFrameErr := godwarf.GetDebugSectionElf(exe, "frame")
-	ehFrameSection := exe.Section(".eh_frame")
+	debugFrameData, debugFrameErr := godwarf.GetDebugSectionElf(dwarfFile, "frame")
+	ehFrameSection := exeFile.Section(".eh_frame")
 	var ehFrameData []byte
 	var ehFrameAddr uint64
 	if ehFrameSection != nil {
@@ -1273,7 +1426,7 @@ func (bi *BinaryInfo) setGStructOffsetElf(image *Image, exe *elf.File, wg *sync.
 
 	switch exe.Machine {
 	case elf.EM_X86_64, elf.EM_386:
-		tlsg := getSymbol(image, exe, "runtime.tlsg")
+		tlsg := getSymbol(image, bi.logger, exe, "runtime.tlsg")
 		if tlsg == nil || tls == nil {
 			bi.gStructOffset = ^uint64(bi.Arch.PtrSize()) + 1 //-ptrSize
 			return
@@ -1290,7 +1443,7 @@ func (bi *BinaryInfo) setGStructOffsetElf(image *Image, exe *elf.File, wg *sync.
 		bi.gStructOffset = ^(memsz) + 1 + tlsg.Value // -tls.Memsz + tlsg.Value
 
 	case elf.EM_AARCH64:
-		tlsg := getSymbol(image, exe, "runtime.tls_g")
+		tlsg := getSymbol(image, bi.logger, exe, "runtime.tls_g")
 		if tlsg == nil || tls == nil {
 			bi.gStructOffset = 2 * uint64(bi.Arch.PtrSize())
 			return
@@ -1304,10 +1457,10 @@ func (bi *BinaryInfo) setGStructOffsetElf(image *Image, exe *elf.File, wg *sync.
 	}
 }
 
-func getSymbol(image *Image, exe *elf.File, name string) *elf.Symbol {
+func getSymbol(image *Image, logger *logrus.Entry, exe *elf.File, name string) *elf.Symbol {
 	symbols, err := exe.Symbols()
 	if err != nil {
-		image.setLoadError("could not parse ELF symbols: %v", err)
+		image.setLoadError(logger, "could not parse ELF symbols: %v", err)
 		return nil
 	}
 
@@ -1366,6 +1519,8 @@ func loadBinaryInfoPE(bi *BinaryInfo, image *Image, path string, entryPoint uint
 	image.loclist5 = loclist.NewDwarf5Reader(debugLoclistBytes)
 	debugAddrBytes, _ := godwarf.GetDebugSectionPE(peFile, "addr")
 	image.debugAddr = godwarf.ParseAddr(debugAddrBytes)
+	debugLineStrBytes, _ := godwarf.GetDebugSectionPE(peFile, "line_str")
+	image.debugLineStr = debugLineStrBytes
 
 	wg.Add(2)
 	go bi.parseDebugFramePE(image, peFile, debugInfoBytes, wg)
@@ -1442,6 +1597,8 @@ func loadBinaryInfoMacho(bi *BinaryInfo, image *Image, path string, entryPoint u
 	image.loclist5 = loclist.NewDwarf5Reader(debugLoclistBytes)
 	debugAddrBytes, _ := godwarf.GetDebugSectionMacho(exe, "addr")
 	image.debugAddr = godwarf.ParseAddr(debugAddrBytes)
+	debugLineStrBytes, _ := godwarf.GetDebugSectionMacho(exe, "line_str")
+	image.debugLineStr = debugLineStrBytes
 
 	wg.Add(2)
 	go bi.parseDebugFrameMacho(image, exe, debugInfoBytes, wg)
@@ -1558,7 +1715,7 @@ func (bi *BinaryInfo) macOSDebugFrameBugWorkaround() {
 func (bi *BinaryInfo) findType(name string) (godwarf.Type, error) {
 	ref, found := bi.types[name]
 	if !found {
-		return nil, reader.TypeNotFoundErr
+		return nil, reader.ErrTypeNotFound
 	}
 	image := bi.Images[ref.imageIndex]
 	return godwarf.ReadType(image.dwarf, ref.imageIndex, ref.offset, image.typeCache)
@@ -1678,6 +1835,9 @@ func (bi *BinaryInfo) loadDebugInfoMaps(image *Image, debugInfoBytes, debugLineB
 	if bi.inlinedCallLines == nil {
 		bi.inlinedCallLines = make(map[fileLine][]uint64)
 	}
+	if bi.dwrapUnwrapCache == nil {
+		bi.dwrapUnwrapCache = make(map[uint64]*Function)
+	}
 
 	image.runtimeTypeToDIE = make(map[uint64]runtimeTypeDIE)
 
@@ -1685,9 +1845,13 @@ func (bi *BinaryInfo) loadDebugInfoMaps(image *Image, debugInfoBytes, debugLineB
 
 	reader := image.DwarfReader()
 
-	for entry, err := reader.Next(); entry != nil; entry, err = reader.Next() {
+	for {
+		entry, err := reader.Next()
 		if err != nil {
-			image.setLoadError("error reading debug_info: %v", err)
+			image.setLoadError(bi.logger, "error reading debug_info: %v", err)
+			break
+		}
+		if entry == nil {
 			break
 		}
 		switch entry.Tag {
@@ -1723,7 +1887,7 @@ func (bi *BinaryInfo) loadDebugInfoMaps(image *Image, debugInfoBytes, debugLineB
 						logger.Printf(fmt, args)
 					}
 				}
-				cu.lineInfo = line.Parse(compdir, bytes.NewBuffer(debugLineBytes[lineInfoOffset:]), logfn, image.StaticBase, bi.GOOS == "windows", bi.Arch.PtrSize())
+				cu.lineInfo = line.Parse(compdir, bytes.NewBuffer(debugLineBytes[lineInfoOffset:]), image.debugLineStr, logfn, image.StaticBase, bi.GOOS == "windows", bi.Arch.PtrSize())
 			}
 			cu.producer, _ = entry.Val(dwarf.AttrProducer).(string)
 			if cu.isgo && cu.producer != "" {
@@ -1732,6 +1896,13 @@ func (bi *BinaryInfo) loadDebugInfoMaps(image *Image, debugInfoBytes, debugLineB
 					cu.optimized = goversion.ProducerAfterOrEqual(cu.producer, 1, 10)
 				} else {
 					cu.optimized = !strings.Contains(cu.producer[semicolon:], "-N") || !strings.Contains(cu.producer[semicolon:], "-l")
+					const regabi = " regabi"
+					if i := strings.Index(cu.producer[semicolon:], regabi); i > 0 {
+						i += semicolon
+						if i+len(regabi) >= len(cu.producer) || cu.producer[i+len(regabi)] == ' ' {
+							bi.regabi = true
+						}
+					}
 					cu.producer = cu.producer[:semicolon]
 				}
 			}
@@ -1758,6 +1929,7 @@ func (bi *BinaryInfo) loadDebugInfoMaps(image *Image, debugInfoBytes, debugLineB
 	sort.Sort(packageVarsByAddr(bi.packageVars))
 
 	bi.LookupFunc = make(map[string]*Function)
+	bi.lookupGenericFunc = nil
 	for i := range bi.Functions {
 		bi.LookupFunc[bi.Functions[i].Name] = &bi.Functions[i]
 	}
@@ -1772,22 +1944,60 @@ func (bi *BinaryInfo) loadDebugInfoMaps(image *Image, debugInfoBytes, debugLineB
 	sort.Strings(bi.Sources)
 	bi.Sources = uniq(bi.Sources)
 
+	if bi.regabi {
+		// prepare patch for runtime.mallocgc's DIE
+		fn := bi.LookupFunc["runtime.mallocgc"]
+		if fn != nil && fn.cu.image == image {
+			tree, err := image.getDwarfTree(fn.offset)
+			if err == nil {
+				tree.Children, err = regabiMallocgcWorkaround(bi)
+				if err != nil {
+					bi.logger.Errorf("could not patch runtime.mallogc: %v", err)
+				} else {
+					image.runtimeMallocgcTree = tree
+				}
+			}
+		}
+	}
+
 	if cont != nil {
 		cont()
 	}
 }
 
+// LookupGenericFunc returns a map that allows searching for instantiations of generic function by specificying a function name without type parameters.
+// For example the key "pkg.(*Receiver).Amethod" will find all instantiations of Amethod:
+//  - pkg.(*Receiver[.shape.int]).Amethod"
+//  - pkg.(*Receiver[.shape.*uint8]).Amethod"
+//  - etc.
+func (bi *BinaryInfo) LookupGenericFunc() map[string][]*Function {
+	if bi.lookupGenericFunc == nil {
+		bi.lookupGenericFunc = make(map[string][]*Function)
+		for i := range bi.Functions {
+			dn := bi.Functions[i].NameWithoutTypeParams()
+			if dn != bi.Functions[i].Name {
+				bi.lookupGenericFunc[dn] = append(bi.lookupGenericFunc[dn], &bi.Functions[i])
+			}
+		}
+	}
+	return bi.lookupGenericFunc
+}
+
 // loadDebugInfoMapsCompileUnit loads entry from a single compile unit.
 func (bi *BinaryInfo) loadDebugInfoMapsCompileUnit(ctxt *loadDebugInfoMapsContext, image *Image, reader *reader.Reader, cu *compileUnit) {
 	hasAttrGoPkgName := goversion.ProducerAfterOrEqual(cu.producer, 1, 13)
 
 	depth := 0
 
-	for entry, err := reader.Next(); entry != nil; entry, err = reader.Next() {
+	for {
+		entry, err := reader.Next()
 		if err != nil {
-			image.setLoadError("error reading debug_info: %v", err)
+			image.setLoadError(bi.logger, "error reading debug_info: %v", err)
 			return
 		}
+		if entry == nil {
+			break
+		}
 		switch entry.Tag {
 		case 0:
 			if depth == 0 {
@@ -1851,7 +2061,7 @@ func (bi *BinaryInfo) loadDebugInfoMapsCompileUnit(ctxt *loadDebugInfoMapsContex
 		case dwarf.TagSubprogram:
 			inlined := false
 			if inval, ok := entry.Val(dwarf.AttrInline).(int64); ok {
-				inlined = inval == 1
+				inlined = inval >= 1
 			}
 
 			if inlined {
@@ -1952,6 +2162,8 @@ func (bi *BinaryInfo) addConcreteSubprogram(entry *dwarf.Entry, ctxt *loadDebugI
 		bi.logger.Warnf("reading debug_info: concrete subprogram without name at %#x", entry.Offset)
 	}
 
+	trampoline, _ := entry.Val(dwarf.AttrTrampoline).(bool)
+
 	originIdx := ctxt.lookupAbstractOrigin(bi, entry.Offset)
 	fn := &bi.Functions[originIdx]
 
@@ -1960,6 +2172,7 @@ func (bi *BinaryInfo) addConcreteSubprogram(entry *dwarf.Entry, ctxt *loadDebugI
 	fn.End = highpc
 	fn.offset = entry.Offset
 	fn.cu = cu
+	fn.trampoline = trampoline
 
 	if entry.Children {
 		bi.loadDebugInfoMapsInlinedCalls(ctxt, reader, cu)
@@ -1991,7 +2204,7 @@ func (bi *BinaryInfo) loadDebugInfoMapsInlinedCalls(ctxt *loadDebugInfoMapsConte
 	for {
 		entry, err := reader.Next()
 		if err != nil {
-			cu.image.setLoadError("error reading debug_info: %v", err)
+			cu.image.setLoadError(bi.logger, "error reading debug_info: %v", err)
 			return
 		}
 		switch entry.Tag {
@@ -2041,6 +2254,10 @@ func (bi *BinaryInfo) loadDebugInfoMapsInlinedCalls(ctxt *loadDebugInfoMapsConte
 
 			fl := fileLine{callfile, int(callline)}
 			bi.inlinedCallLines[fl] = append(bi.inlinedCallLines[fl], lowpc)
+
+			if entry.Children {
+				bi.loadDebugInfoMapsInlinedCalls(ctxt, reader, cu)
+			}
 		}
 		reader.SkipChildren()
 	}
