@@ -1,6 +1,7 @@
 package proc
 
 import (
+	"debug/dwarf"
 	"errors"
 	"fmt"
 	"go/ast"
@@ -8,6 +9,12 @@ import (
 	"go/parser"
 	"go/token"
 	"reflect"
+
+	"github.com/go-delve/delve/pkg/dwarf/godwarf"
+	"github.com/go-delve/delve/pkg/dwarf/op"
+	"github.com/go-delve/delve/pkg/dwarf/reader"
+	"github.com/go-delve/delve/pkg/goversion"
+	"github.com/go-delve/delve/pkg/proc/internal/ebpf"
 )
 
 const (
@@ -20,6 +27,8 @@ const (
 
 	unrecoveredPanicID = -1
 	fatalThrowID       = -2
+
+	NoLogicalID = -1000 // Logical breakpoint ID for breakpoints internal breakpoints.
 )
 
 // Breakpoint represents a physical breakpoint. Stores information on the break
@@ -34,27 +43,43 @@ type Breakpoint struct {
 	Addr         uint64 // Address breakpoint is set for.
 	OriginalData []byte // If software breakpoint, the data we replace with breakpoint instruction.
 	Name         string // User defined name of the breakpoint
-	LogicalID    int    // ID of the logical breakpoint that owns this physical breakpoint
 
-	WatchExpr    string
-	WatchType    WatchType
-	HWBreakIndex uint8 // hardware breakpoint index
+	WatchExpr     string
+	WatchType     WatchType
+	HWBreakIndex  uint8 // hardware breakpoint index
+	watchStackOff int64 // for watchpoints of stack variables, offset of the address from top of the stack
 
-	// Kind describes whether this is an internal breakpoint (for next'ing or
-	// stepping).
-	// A single breakpoint can be both a UserBreakpoint and some kind of
-	// internal breakpoint, but it can not be two different kinds of internal
-	// breakpoint.
-	Kind BreakpointKind
+	// Breaklets is the list of overlapping breakpoints on this physical breakpoint.
+	// There can be at most one UserBreakpoint in this list but multiple internal breakpoints are allowed.
+	Breaklets []*Breaklet
 
 	// Breakpoint information
-	Tracepoint    bool // Tracepoint flag
-	TraceReturn   bool
-	Goroutine     bool     // Retrieve goroutine information
-	Stacktrace    int      // Number of stack frames to retrieve
-	Variables     []string // Variables to evaluate
-	LoadArgs      *LoadConfig
-	LoadLocals    *LoadConfig
+	Tracepoint  bool // Tracepoint flag
+	TraceReturn bool
+	Goroutine   bool     // Retrieve goroutine information
+	Stacktrace  int      // Number of stack frames to retrieve
+	Variables   []string // Variables to evaluate
+	LoadArgs    *LoadConfig
+	LoadLocals  *LoadConfig
+	UserData    interface{} // Any additional information about the breakpoint
+
+	// ReturnInfo describes how to collect return variables when this
+	// breakpoint is hit as a return breakpoint.
+	returnInfo *returnBreakpointInfo
+}
+
+// Breaklet represents one of multiple breakpoints that can overlap on a
+// single physical breakpoint.
+type Breaklet struct {
+	// Kind describes whether this is a stepping breakpoint (for next'ing or
+	// stepping).
+	Kind BreakpointKind
+
+	LogicalID int // ID of the logical breakpoint that owns this physical breakpoint
+
+	// Cond: if not nil the breakpoint will be triggered only if evaluating Cond returns true
+	Cond ast.Expr
+
 	HitCount      map[int]uint64 // Number of times a breakpoint has been reached in a certain goroutine
 	TotalHitCount uint64         // Number of times a breakpoint has been reached
 
@@ -68,10 +93,7 @@ type Breakpoint struct {
 	// function only triggers on panic or on the defer call to
 	// the function, not when the function is called directly
 	DeferReturns []uint64
-	// Cond: if not nil the breakpoint will be triggered only if evaluating Cond returns true
-	Cond ast.Expr
-	// internalCond is the same as Cond but used for the condition of internal breakpoints
-	internalCond ast.Expr
+
 	// HitCond: if not nil the breakpoint will be triggered only if the evaluated HitCond returns
 	// true with the TotalHitCount.
 	HitCond *struct {
@@ -79,9 +101,19 @@ type Breakpoint struct {
 		Val int
 	}
 
-	// ReturnInfo describes how to collect return variables when this
-	// breakpoint is hit as a return breakpoint.
-	returnInfo *returnBreakpointInfo
+	// checkPanicCall checks that the breakpoint happened while the function was
+	// called by a panic. It is only checked for WatchOutOfScopeBreakpoint Kind.
+	checkPanicCall bool
+
+	// callback is called if every other condition for this breaklet is met,
+	// the return value will determine if the breaklet should be considered
+	// active.
+	// The callback can have side-effects.
+	callback func(th Thread) bool
+
+	// For WatchOutOfScopeBreakpoints and StackResizeBreakpoints the watchpoint
+	// field contains the watchpoint related to this out of scope sentinel.
+	watchpoint *Breakpoint
 }
 
 // BreakpointKind determines the behavior of delve when the
@@ -103,6 +135,16 @@ const (
 	// Continue will set a new breakpoint (of NextBreakpoint kind) on the
 	// destination of CALL, delete this breakpoint and then continue again
 	StepBreakpoint
+
+	// WatchOutOfScopeBreakpoint is a breakpoint used to detect when a watched
+	// stack variable goes out of scope.
+	WatchOutOfScopeBreakpoint
+
+	// StackResizeBreakpoint is a breakpoint used to detect stack resizes to
+	// adjust the watchpoint of stack variables.
+	StackResizeBreakpoint
+
+	steppingMask = NextBreakpoint | NextDeferBreakpoint | StepBreakpoint
 )
 
 // WatchType is the watchpoint type
@@ -136,7 +178,48 @@ func (wtype WatchType) withSize(sz uint8) WatchType {
 var ErrHWBreakUnsupported = errors.New("hardware breakpoints not implemented")
 
 func (bp *Breakpoint) String() string {
-	return fmt.Sprintf("Breakpoint %d at %#v %s:%d (%d)", bp.LogicalID, bp.Addr, bp.File, bp.Line, bp.TotalHitCount)
+	return fmt.Sprintf("Breakpoint %d at %#v %s:%d", bp.LogicalID(), bp.Addr, bp.File, bp.Line)
+}
+
+func (bp *Breakpoint) LogicalID() int {
+	for _, breaklet := range bp.Breaklets {
+		if breaklet.Kind == UserBreakpoint {
+			return breaklet.LogicalID
+		}
+	}
+	return NoLogicalID
+}
+
+// VerboseDescr returns a string describing parts of the breakpoint struct
+// that aren't otherwise user visible, for debugging purposes.
+func (bp *Breakpoint) VerboseDescr() []string {
+	r := []string{}
+
+	r = append(r, fmt.Sprintf("OriginalData=%#x", bp.OriginalData))
+
+	if bp.WatchType != 0 {
+		r = append(r, fmt.Sprintf("HWBreakIndex=%#x watchStackOff=%#x", bp.HWBreakIndex, bp.watchStackOff))
+	}
+
+	for _, breaklet := range bp.Breaklets {
+		switch breaklet.Kind {
+		case UserBreakpoint:
+			r = append(r, fmt.Sprintf("User Cond=%q HitCond=%v", exprToString(breaklet.Cond), breaklet.HitCond))
+		case NextBreakpoint:
+			r = append(r, fmt.Sprintf("Next Cond=%q", exprToString(breaklet.Cond)))
+		case NextDeferBreakpoint:
+			r = append(r, fmt.Sprintf("NextDefer Cond=%q DeferReturns=%#x", exprToString(breaklet.Cond), breaklet.DeferReturns))
+		case StepBreakpoint:
+			r = append(r, fmt.Sprintf("Step Cond=%q", exprToString(breaklet.Cond)))
+		case WatchOutOfScopeBreakpoint:
+			r = append(r, fmt.Sprintf("WatchOutOfScope Cond=%q checkPanicCall=%v", exprToString(breaklet.Cond), breaklet.checkPanicCall))
+		case StackResizeBreakpoint:
+			r = append(r, fmt.Sprintf("StackResizeBreakpoint Cond=%q", exprToString(breaklet.Cond)))
+		default:
+			r = append(r, fmt.Sprintf("Unknown %d", breaklet.Kind))
+		}
+	}
+	return r
 }
 
 // BreakpointExistsError is returned when trying to set a breakpoint at
@@ -169,74 +252,102 @@ type returnBreakpointInfo struct {
 }
 
 // CheckCondition evaluates bp's condition on thread.
-func (bp *Breakpoint) CheckCondition(thread Thread) BreakpointState {
-	bpstate := BreakpointState{Breakpoint: bp, Active: false, Internal: false, CondError: nil}
-	bpstate.checkCond(thread)
-	// Update the breakpoint hit counts.
-	if bpstate.Breakpoint != nil && bpstate.Active {
-		if g, err := GetG(thread); err == nil {
-			bpstate.HitCount[g.ID]++
-		}
-		bpstate.TotalHitCount++
+func (bp *Breakpoint) checkCondition(tgt *Target, thread Thread, bpstate *BreakpointState) {
+	*bpstate = BreakpointState{Breakpoint: bp, Active: false, Stepping: false, SteppingInto: false, CondError: nil}
+	for _, breaklet := range bp.Breaklets {
+		bpstate.checkCond(tgt, breaklet, thread)
 	}
-	bpstate.checkHitCond(thread)
-	return bpstate
 }
 
-func (bpstate *BreakpointState) checkCond(thread Thread) {
-	if bpstate.Cond == nil && bpstate.internalCond == nil {
-		bpstate.Active = true
-		bpstate.Internal = bpstate.IsInternal()
+func (bpstate *BreakpointState) checkCond(tgt *Target, breaklet *Breaklet, thread Thread) {
+	var condErr error
+	active := true
+	if breaklet.Cond != nil {
+		active, condErr = evalBreakpointCondition(tgt, thread, breaklet.Cond)
+	}
+
+	if condErr != nil && bpstate.CondError == nil {
+		bpstate.CondError = condErr
+	}
+	if !active {
 		return
 	}
-	nextDeferOk := true
-	if bpstate.Kind&NextDeferBreakpoint != 0 {
-		var err error
-		frames, err := ThreadStacktrace(thread, 2)
-		if err == nil {
-			nextDeferOk, _ = isPanicCall(frames)
-			if !nextDeferOk {
-				nextDeferOk, _ = isDeferReturnCall(frames, bpstate.DeferReturns)
+
+	switch breaklet.Kind {
+	case UserBreakpoint:
+		if g, err := GetG(thread); err == nil {
+			breaklet.HitCount[g.ID]++
+		}
+		breaklet.TotalHitCount++
+		active = checkHitCond(breaklet)
+
+	case StepBreakpoint, NextBreakpoint, NextDeferBreakpoint:
+		nextDeferOk := true
+		if breaklet.Kind&NextDeferBreakpoint != 0 {
+			var err error
+			frames, err := ThreadStacktrace(thread, 2)
+			if err == nil {
+				nextDeferOk, _ = isPanicCall(frames)
+				if !nextDeferOk {
+					nextDeferOk, _ = isDeferReturnCall(frames, breaklet.DeferReturns)
+				}
 			}
 		}
-	}
-	if bpstate.IsInternal() {
-		// Check internalCondition if this is also an internal breakpoint
-		bpstate.Active, bpstate.CondError = evalBreakpointCondition(thread, bpstate.internalCond)
-		bpstate.Active = bpstate.Active && nextDeferOk
-		if bpstate.Active || bpstate.CondError != nil {
-			bpstate.Internal = true
-			return
+		active = active && nextDeferOk
+		if active {
+			bpstate.Stepping = true
+			if breaklet.Kind == StepBreakpoint {
+				bpstate.SteppingInto = true
+			}
 		}
+
+	case WatchOutOfScopeBreakpoint:
+		if breaklet.checkPanicCall {
+			frames, err := ThreadStacktrace(thread, 2)
+			if err == nil {
+				ipc, _ := isPanicCall(frames)
+				active = active && ipc
+			}
+		}
+
+	case StackResizeBreakpoint:
+		// no further checks
+
+	default:
+		bpstate.CondError = fmt.Errorf("internal error unknown breakpoint kind %v", breaklet.Kind)
 	}
-	if bpstate.IsUser() {
-		// Check normal condition if this is also a user breakpoint
-		bpstate.Active, bpstate.CondError = evalBreakpointCondition(thread, bpstate.Cond)
+
+	if active {
+		if breaklet.callback != nil {
+			active = breaklet.callback(thread)
+		}
+		bpstate.Active = active
 	}
 }
 
 // checkHitCond evaluates bp's hit condition on thread.
-func (bpstate *BreakpointState) checkHitCond(thread Thread) {
-	if bpstate.HitCond == nil || !bpstate.Active || bpstate.Internal {
-		return
+func checkHitCond(breaklet *Breaklet) bool {
+	if breaklet.HitCond == nil {
+		return true
 	}
 	// Evaluate the breakpoint condition.
-	switch bpstate.HitCond.Op {
+	switch breaklet.HitCond.Op {
 	case token.EQL:
-		bpstate.Active = int(bpstate.TotalHitCount) == bpstate.HitCond.Val
+		return int(breaklet.TotalHitCount) == breaklet.HitCond.Val
 	case token.NEQ:
-		bpstate.Active = int(bpstate.TotalHitCount) != bpstate.HitCond.Val
+		return int(breaklet.TotalHitCount) != breaklet.HitCond.Val
 	case token.GTR:
-		bpstate.Active = int(bpstate.TotalHitCount) > bpstate.HitCond.Val
+		return int(breaklet.TotalHitCount) > breaklet.HitCond.Val
 	case token.LSS:
-		bpstate.Active = int(bpstate.TotalHitCount) < bpstate.HitCond.Val
+		return int(breaklet.TotalHitCount) < breaklet.HitCond.Val
 	case token.GEQ:
-		bpstate.Active = int(bpstate.TotalHitCount) >= bpstate.HitCond.Val
+		return int(breaklet.TotalHitCount) >= breaklet.HitCond.Val
 	case token.LEQ:
-		bpstate.Active = int(bpstate.TotalHitCount) <= bpstate.HitCond.Val
+		return int(breaklet.TotalHitCount) <= breaklet.HitCond.Val
 	case token.REM:
-		bpstate.Active = int(bpstate.TotalHitCount)%bpstate.HitCond.Val == 0
+		return int(breaklet.TotalHitCount)%breaklet.HitCond.Val == 0
 	}
+	return false
 }
 
 func isPanicCall(frames []Stackframe) (bool, int) {
@@ -261,6 +372,13 @@ func isPanicCall(frames []Stackframe) (bool, int) {
 }
 
 func isDeferReturnCall(frames []Stackframe, deferReturns []uint64) (bool, uint64) {
+	if len(frames) >= 2 && (len(deferReturns) > 0) {
+		// On Go 1.18 and later runtime.deferreturn doesn't use jmpdefer anymore,
+		// it's a normal function making normal calls to deferred functions.
+		if frames[1].Current.Fn != nil && frames[1].Current.Fn.Name == "runtime.deferreturn" {
+			return true, 0
+		}
+	}
 	if len(frames) >= 1 {
 		for _, pc := range deferReturns {
 			if frames[0].Ret == pc {
@@ -271,27 +389,48 @@ func isDeferReturnCall(frames []Stackframe, deferReturns []uint64) (bool, uint64
 	return false, 0
 }
 
-// IsInternal returns true if bp is an internal breakpoint.
-// User-set breakpoints can overlap with internal breakpoints, in that case
-// both IsUser and IsInternal will be true.
-func (bp *Breakpoint) IsInternal() bool {
-	return bp.Kind != UserBreakpoint
+// IsStepping returns true if bp is an stepping breakpoint.
+// User-set breakpoints can overlap with stepping breakpoints, in that case
+// both IsUser and IsStepping will be true.
+func (bp *Breakpoint) IsStepping() bool {
+	for _, breaklet := range bp.Breaklets {
+		if breaklet.Kind&steppingMask != 0 {
+			return true
+		}
+	}
+	return false
 }
 
 // IsUser returns true if bp is a user-set breakpoint.
-// User-set breakpoints can overlap with internal breakpoints, in that case
-// both IsUser and IsInternal will be true.
+// User-set breakpoints can overlap with stepping breakpoints, in that case
+// both IsUser and IsStepping will be true.
 func (bp *Breakpoint) IsUser() bool {
-	return bp.Kind&UserBreakpoint != 0
+	for _, breaklet := range bp.Breaklets {
+		if breaklet.Kind == UserBreakpoint {
+			return true
+		}
+	}
+	return false
 }
 
-func evalBreakpointCondition(thread Thread, cond ast.Expr) (bool, error) {
+// UserBreaklet returns the user breaklet for this breakpoint, or nil if
+// none exist.
+func (bp *Breakpoint) UserBreaklet() *Breaklet {
+	for _, breaklet := range bp.Breaklets {
+		if breaklet.Kind == UserBreakpoint {
+			return breaklet
+		}
+	}
+	return nil
+}
+
+func evalBreakpointCondition(tgt *Target, thread Thread, cond ast.Expr) (bool, error) {
 	if cond == nil {
 		return true, nil
 	}
-	scope, err := GoroutineScope(nil, thread)
+	scope, err := GoroutineScope(tgt, thread)
 	if err != nil {
-		scope, err = ThreadScope(nil, thread)
+		scope, err = ThreadScope(tgt, thread)
 		if err != nil {
 			return true, err
 		}
@@ -324,8 +463,11 @@ func (nbp NoBreakpointError) Error() string {
 type BreakpointMap struct {
 	M map[uint64]*Breakpoint
 
-	breakpointIDCounter         int
-	internalBreakpointIDCounter int
+	// WatchOutOfScope is the list of watchpoints that went out of scope during
+	// the last resume operation
+	WatchOutOfScope []*Breakpoint
+
+	breakpointIDCounter int
 }
 
 // NewBreakpointMap creates a new BreakpointMap.
@@ -341,6 +483,103 @@ func (t *Target) SetBreakpoint(addr uint64, kind BreakpointKind, cond ast.Expr)
 	return t.setBreakpointInternal(addr, kind, 0, cond)
 }
 
+// SetEBPFTracepoint will attach a uprobe to the function
+// specified by 'fnName'.
+func (t *Target) SetEBPFTracepoint(fnName string) error {
+	// Not every OS/arch that we support has support for eBPF,
+	// so check early and return an error if this is called on an
+	// unsupported system.
+	if !t.proc.SupportsBPF() {
+		return errors.New("eBPF is not supported")
+	}
+	fns, err := t.BinInfo().FindFunction(fnName)
+	if err != nil {
+		return err
+	}
+
+	// Get information on the Goroutine so we can tell the
+	// eBPF program where to find it in order to get the
+	// goroutine ID.
+	rdr := t.BinInfo().Images[0].DwarfReader()
+	rdr.SeekToTypeNamed("runtime.g")
+	typ, err := t.BinInfo().findType("runtime.g")
+	if err != nil {
+		return errors.New("could not find type for runtime.g")
+	}
+	var goidOffset int64
+	switch t := typ.(type) {
+	case *godwarf.StructType:
+		for _, field := range t.Field {
+			if field.Name == "goid" {
+				goidOffset = field.ByteOffset
+				break
+			}
+		}
+	}
+
+	for _, fn := range fns {
+		err := t.setEBPFTracepointOnFunc(fn, goidOffset)
+		if err != nil {
+			return err
+		}
+	}
+	return nil
+}
+
+func (t *Target) setEBPFTracepointOnFunc(fn *Function, goidOffset int64) error {
+	// Start putting together the argument map. This will tell the eBPF program
+	// all of the arguments we want to trace and how to find them.
+
+	// Start looping through each argument / return parameter for the function we
+	// are setting the uprobe on. Parse location information so that we can pass it
+	// along to the eBPF program.
+	dwarfTree, err := fn.cu.image.getDwarfTree(fn.offset)
+	if err != nil {
+		return err
+	}
+	variablesFlags := reader.VariablesOnlyVisible
+	if t.BinInfo().Producer() != "" && goversion.ProducerAfterOrEqual(t.BinInfo().Producer(), 1, 15) {
+		variablesFlags |= reader.VariablesTrustDeclLine
+	}
+	_, l, _ := t.BinInfo().PCToLine(fn.Entry)
+
+	var args []ebpf.UProbeArgMap
+	varEntries := reader.Variables(dwarfTree, fn.Entry, l, variablesFlags)
+	for _, entry := range varEntries {
+		_, dt, err := readVarEntry(entry.Tree, fn.cu.image)
+		if err != nil {
+			return err
+		}
+
+		offset, pieces, _, err := t.BinInfo().Location(entry, dwarf.AttrLocation, fn.Entry, op.DwarfRegisters{}, nil)
+		if err != nil {
+			return err
+		}
+		paramPieces := make([]int, 0, len(pieces))
+		for _, piece := range pieces {
+			if piece.Kind == op.RegPiece {
+				paramPieces = append(paramPieces, int(piece.Val))
+			}
+		}
+		isret, _ := entry.Val(dwarf.AttrVarParam).(bool)
+		offset += int64(t.BinInfo().Arch.PtrSize())
+		args = append(args, ebpf.UProbeArgMap{
+			Offset: offset,
+			Size:   dt.Size(),
+			Kind:   dt.Common().ReflectKind,
+			Pieces: paramPieces,
+			InReg:  len(pieces) > 0,
+			Ret:    isret,
+		})
+	}
+
+	//TODO(aarzilli): inlined calls?
+
+	// Finally, set the uprobe on the function.
+	t.proc.SetUProbe(fn.Name, goidOffset, args)
+	return nil
+}
+
 // SetWatchpoint sets a data breakpoint at addr and stores it in the
 // process wide break point table.
 func (t *Target) SetWatchpoint(scope *EvalScope, expr string, wtype WatchType, cond ast.Expr) (*Breakpoint, error) {
@@ -372,36 +611,61 @@ func (t *Target) SetWatchpoint(scope *EvalScope, expr string, wtype WatchType, c
 		//member fields here.
 		return nil, fmt.Errorf("can not watch variable of type %s", xv.DwarfType.String())
 	}
-	if xv.Addr >= scope.g.stack.lo && xv.Addr < scope.g.stack.hi {
-		//TODO(aarzilli): support watching stack variables
-		return nil, errors.New("can not watch stack allocated variable")
+
+	stackWatch := scope.g != nil && !scope.g.SystemStack && xv.Addr >= scope.g.stack.lo && xv.Addr < scope.g.stack.hi
+
+	if stackWatch && wtype&WatchRead != 0 {
+		// In theory this would work except for the fact that the runtime will
+		// read them randomly to resize stacks so it doesn't make sense to do
+		// this.
+		return nil, errors.New("can not watch stack allocated variable for reads")
 	}
 
 	bp, err := t.setBreakpointInternal(xv.Addr, UserBreakpoint, wtype.withSize(uint8(sz)), cond)
-	if bp != nil {
-		bp.WatchExpr = expr
+	if err != nil {
+		return bp, err
 	}
-	return bp, err
+	bp.WatchExpr = expr
+
+	if stackWatch {
+		bp.watchStackOff = int64(bp.Addr) - int64(scope.g.stack.hi)
+		err := t.setStackWatchBreakpoints(scope, bp)
+		if err != nil {
+			return bp, err
+		}
+	}
+
+	return bp, nil
 }
 
 func (t *Target) setBreakpointInternal(addr uint64, kind BreakpointKind, wtype WatchType, cond ast.Expr) (*Breakpoint, error) {
 	if valid, err := t.Valid(); !valid {
-		return nil, err
+		recorded, _ := t.Recorded()
+		if !recorded {
+			return nil, err
+		}
 	}
 	bpmap := t.Breakpoints()
+	newBreaklet := &Breaklet{Kind: kind, Cond: cond}
+	if kind == UserBreakpoint {
+		newBreaklet.HitCount = map[int]uint64{}
+		bpmap.breakpointIDCounter++
+		newBreaklet.LogicalID = bpmap.breakpointIDCounter
+	}
 	if bp, ok := bpmap.M[addr]; ok {
-		// We can overlap one internal breakpoint with one user breakpoint, we
-		// need to support this otherwise a conditional breakpoint can mask a
-		// breakpoint set by next or step.
-		if (kind != UserBreakpoint && bp.Kind != UserBreakpoint) || (kind == UserBreakpoint && bp.IsUser()) {
+		if !bp.canOverlap(kind) {
 			return bp, BreakpointExistsError{bp.File, bp.Line, bp.Addr}
 		}
-		bp.Kind |= kind
-		if kind != UserBreakpoint {
-			bp.internalCond = cond
-		} else {
-			bp.Cond = cond
+		if kind == UserBreakpoint {
+			bp.Tracepoint = false
+			bp.TraceReturn = false
+			bp.Goroutine = false
+			bp.Stacktrace = 0
+			bp.Variables = nil
+			bp.LoadArgs = nil
+			bp.LoadLocals = nil
 		}
+		bp.Breaklets = append(bp.Breaklets, newBreaklet)
 		return bp, nil
 	}
 
@@ -434,8 +698,6 @@ func (t *Target) setBreakpointInternal(addr uint64, kind BreakpointKind, wtype W
 		File:         f,
 		Line:         l,
 		Addr:         addr,
-		Kind:         kind,
-		HitCount:     map[int]uint64{},
 	}
 
 	err := t.proc.WriteBreakpoint(newBreakpoint)
@@ -443,15 +705,7 @@ func (t *Target) setBreakpointInternal(addr uint64, kind BreakpointKind, wtype W
 		return nil, err
 	}
 
-	if kind != UserBreakpoint {
-		bpmap.internalBreakpointIDCounter++
-		newBreakpoint.LogicalID = bpmap.internalBreakpointIDCounter
-		newBreakpoint.internalCond = cond
-	} else {
-		bpmap.breakpointIDCounter++
-		newBreakpoint.LogicalID = bpmap.breakpointIDCounter
-		newBreakpoint.Cond = cond
-	}
+	newBreakpoint.Breaklets = append(newBreakpoint.Breaklets, newBreaklet)
 
 	bpmap.M[addr] = newBreakpoint
 
@@ -463,68 +717,116 @@ func (t *Target) SetBreakpointWithID(id int, addr uint64) (*Breakpoint, error) {
 	bpmap := t.Breakpoints()
 	bp, err := t.SetBreakpoint(addr, UserBreakpoint, nil)
 	if err == nil {
-		bp.LogicalID = id
-		bpmap.breakpointIDCounter--
+		for _, breaklet := range bp.Breaklets {
+			if breaklet.Kind == UserBreakpoint {
+				breaklet.LogicalID = id
+				bpmap.breakpointIDCounter--
+				break
+			}
+		}
 	}
 	return bp, err
 }
 
-// ClearBreakpoint clears the breakpoint at addr.
-func (t *Target) ClearBreakpoint(addr uint64) (*Breakpoint, error) {
-	if valid, err := t.Valid(); !valid {
-		return nil, err
+// canOverlap returns true if a breakpoint of kind can be overlapped to the
+// already existing breaklets in bp.
+// At most one user breakpoint can be set but multiple internal breakpoints are allowed.
+// All other internal breakpoints are allowed to overlap freely.
+func (bp *Breakpoint) canOverlap(kind BreakpointKind) bool {
+	if kind == UserBreakpoint {
+		return !bp.IsUser()
 	}
-	bpmap := t.Breakpoints()
-	bp, ok := bpmap.M[addr]
-	if !ok {
-		return nil, NoBreakpointError{Addr: addr}
-	}
-
-	bp.Kind &= ^UserBreakpoint
-	bp.Cond = nil
-	if bp.Kind != 0 {
-		return bp, nil
-	}
-
-	if err := t.proc.EraseBreakpoint(bp); err != nil {
-		return nil, err
-	}
-
-	delete(bpmap.M, addr)
-
-	return bp, nil
+	return true
 }
 
-// ClearInternalBreakpoints removes all internal breakpoints from the map,
-// calling clearBreakpoint on each one.
-func (t *Target) ClearInternalBreakpoints() error {
-	bpmap := t.Breakpoints()
-	threads := t.ThreadList()
-	for addr, bp := range bpmap.M {
-		bp.Kind = bp.Kind & UserBreakpoint
-		bp.internalCond = nil
-		bp.returnInfo = nil
-		if bp.Kind != 0 {
-			continue
-		}
-		if err := t.proc.EraseBreakpoint(bp); err != nil {
+// ClearBreakpoint clears the breakpoint at addr.
+func (t *Target) ClearBreakpoint(addr uint64) error {
+	if valid, err := t.Valid(); !valid {
+		recorded, _ := t.Recorded()
+		if !recorded {
 			return err
 		}
-		for _, thread := range threads {
-			if thread.Breakpoint().Breakpoint == bp {
-				thread.Breakpoint().Clear()
+	}
+	bp, ok := t.Breakpoints().M[addr]
+	if !ok {
+		return NoBreakpointError{Addr: addr}
+	}
+
+	for i := range bp.Breaklets {
+		if bp.Breaklets[i].Kind == UserBreakpoint {
+			bp.Breaklets[i] = nil
+		}
+	}
+
+	_, err := t.finishClearBreakpoint(bp)
+	if err != nil {
+		return err
+	}
+
+	if bp.WatchExpr != "" && bp.watchStackOff != 0 {
+		// stack watchpoint, must remove all its WatchOutOfScopeBreakpoints/StackResizeBreakpoints
+		err := t.clearStackWatchBreakpoints(bp)
+		if err != nil {
+			return err
+		}
+	}
+
+	return nil
+}
+
+// ClearSteppingBreakpoints removes all stepping breakpoints from the map,
+// calling clearBreakpoint on each one.
+func (t *Target) ClearSteppingBreakpoints() error {
+	bpmap := t.Breakpoints()
+	threads := t.ThreadList()
+	for _, bp := range bpmap.M {
+		for i := range bp.Breaklets {
+			if bp.Breaklets[i].Kind&steppingMask != 0 {
+				bp.Breaklets[i] = nil
+			}
+		}
+		cleared, err := t.finishClearBreakpoint(bp)
+		if err != nil {
+			return err
+		}
+		if cleared {
+			for _, thread := range threads {
+				if thread.Breakpoint().Breakpoint == bp {
+					thread.Breakpoint().Clear()
+				}
 			}
 		}
-		delete(bpmap.M, addr)
 	}
 	return nil
 }
 
-// HasInternalBreakpoints returns true if bpmap has at least one internal
+// finishClearBreakpoint clears nil breaklets from the breaklet list of bp
+// and if it is empty erases the breakpoint.
+// Returns true if the breakpoint was deleted
+func (t *Target) finishClearBreakpoint(bp *Breakpoint) (bool, error) {
+	oldBreaklets := bp.Breaklets
+	bp.Breaklets = bp.Breaklets[:0]
+	for _, breaklet := range oldBreaklets {
+		if breaklet != nil {
+			bp.Breaklets = append(bp.Breaklets, breaklet)
+		}
+	}
+	if len(bp.Breaklets) > 0 {
+		return false, nil
+	}
+	if err := t.proc.EraseBreakpoint(bp); err != nil {
+		return false, err
+	}
+
+	delete(t.Breakpoints().M, bp.Addr)
+	return true, nil
+}
+
+// HasSteppingBreakpoints returns true if bpmap has at least one stepping
 // breakpoint set.
-func (bpmap *BreakpointMap) HasInternalBreakpoints() bool {
+func (bpmap *BreakpointMap) HasSteppingBreakpoints() bool {
 	for _, bp := range bpmap.M {
-		if bp.IsInternal() {
+		if bp.IsStepping() {
 			return true
 		}
 	}
@@ -544,11 +846,14 @@ func (bpmap *BreakpointMap) HasHWBreakpoints() bool {
 // BreakpointState describes the state of a breakpoint in a thread.
 type BreakpointState struct {
 	*Breakpoint
-	// Active is true if the breakpoint condition was met.
+	// Active is true if the condition of any breaklet is met.
 	Active bool
-	// Internal is true if the breakpoint was matched as an internal
+	// Stepping is true if one of the active breaklets is a stepping
 	// breakpoint.
-	Internal bool
+	Stepping bool
+	// SteppingInto is true if one of the active stepping breaklets has Kind ==
+	// StepBreakpoint.
+	SteppingInto bool
 	// CondError contains any error encountered while evaluating the
 	// breakpoint's condition.
 	CondError error
@@ -558,7 +863,8 @@ type BreakpointState struct {
 func (bpstate *BreakpointState) Clear() {
 	bpstate.Breakpoint = nil
 	bpstate.Active = false
-	bpstate.Internal = false
+	bpstate.Stepping = false
+	bpstate.SteppingInto = false
 	bpstate.CondError = nil
 }
 
@@ -567,8 +873,8 @@ func (bpstate *BreakpointState) String() string {
 	if bpstate.Active {
 		s += " active"
 	}
-	if bpstate.Internal {
-		s += " internal"
+	if bpstate.Stepping {
+		s += " stepping"
 	}
 	return s
 }
@@ -617,7 +923,7 @@ func (rbpi *returnBreakpointInfo) Collect(t *Target, thread Thread) []*Variable
 		return returnInfoError("could not read function entry", err, thread.ProcessMemory())
 	}
 
-	vars, err := scope.Locals()
+	vars, err := scope.Locals(0)
 	if err != nil {
 		return returnInfoError("could not evaluate return variables", err, thread.ProcessMemory())
 	}
