@@ -11,7 +11,7 @@
 // The protocol is specified at:
 //   https://sourceware.org/gdb/onlinedocs/gdb/Remote-Protocol.html
 // with additional documentation for lldb specific extensions described at:
-//   https://github.com/llvm-mirror/lldb/blob/master/docs/lldb-gdb-remote.txt
+//   https://github.com/llvm/llvm-project/blob/main/lldb/docs/lldb-gdb-remote.txt
 //
 // Terminology:
 //  * inferior: the program we are trying to debug
@@ -81,7 +81,9 @@ import (
 	"github.com/go-delve/delve/pkg/elfwriter"
 	"github.com/go-delve/delve/pkg/logflags"
 	"github.com/go-delve/delve/pkg/proc"
+	"github.com/go-delve/delve/pkg/proc/internal/ebpf"
 	"github.com/go-delve/delve/pkg/proc/linutil"
+	"github.com/go-delve/delve/pkg/proc/macutil"
 	isatty "github.com/mattn/go-isatty"
 )
 
@@ -129,6 +131,10 @@ var debugserverExecutablePaths = []string{
 // while there are still internal breakpoints set.
 var ErrDirChange = errors.New("direction change with internal breakpoints")
 
+// ErrStartCallInjectionBackwards is returned when trying to start a call
+// injection while the recording is being run backwards.
+var ErrStartCallInjectionBackwards = errors.New("can not start a call injection while running backwards")
+
 var checkCanUnmaskSignalsOnce sync.Once
 var canUnmaskSignalsCached bool
 
@@ -143,13 +149,15 @@ type gdbProcess struct {
 	currentThread *gdbThread
 
 	exited, detached bool
+	almostExited     bool // true if 'rr' has sent its synthetic SIGKILL
 	ctrlC            bool // ctrl-c was sent to stop inferior
 
 	manualStopRequested bool
 
 	breakpoints proc.BreakpointMap
 
-	gcmdok         bool   // true if the stub supports g and G commands
+	gcmdok         bool   // true if the stub supports g and (maybe) G commands
+	_Gcmdok        bool   // true if the stub supports G command
 	threadStopInfo bool   // true if the stub supports qThreadStopInfo
 	tracedir       string // if attached to rr the path to the trace directory
 
@@ -163,7 +171,7 @@ type gdbProcess struct {
 	onDetach func() // called after a successful detach
 }
 
-var _ proc.ProcessInternal = &gdbProcess{}
+var _ proc.RecordingManipulationInternal = &gdbProcess{}
 
 // gdbThread represents an operating system thread.
 type gdbThread struct {
@@ -172,8 +180,9 @@ type gdbThread struct {
 	regs              gdbRegisters
 	CurrentBreakpoint proc.BreakpointState
 	p                 *gdbProcess
-	sig               uint8 // signal received by thread after last stop
-	setbp             bool  // thread was stopped because of a breakpoint
+	sig               uint8  // signal received by thread after last stop
+	setbp             bool   // thread was stopped because of a breakpoint
+	watchAddr         uint64 // if > 0 this is the watchpoint address
 	common            proc.CommonThread
 }
 
@@ -202,6 +211,8 @@ type gdbRegisters struct {
 type gdbRegister struct {
 	value  []byte
 	regnum int
+
+	ignoreOnWrite bool
 }
 
 // gdbRegname records names of important CPU registers
@@ -221,6 +232,8 @@ func newProcess(process *os.Process) *gdbProcess {
 			inbuf:               make([]byte, 0, initialInputBufferSize),
 			direction:           proc.Forward,
 			log:                 logger,
+			goarch:              runtime.GOARCH,
+			goos:                runtime.GOOS,
 		},
 		threads:        make(map[int]*gdbThread),
 		bi:             proc.NewBinaryInfo(runtime.GOOS, runtime.GOARCH),
@@ -319,26 +332,19 @@ func (p *gdbProcess) Connect(conn net.Conn, path string, pid int, debugInfoDirs
 		return nil, err
 	}
 
-	if verbuf, err := p.conn.exec([]byte("$qGDBServerVersion"), "init"); err == nil {
-		for _, v := range strings.Split(string(verbuf), ";") {
-			if strings.HasPrefix(v, "version:") {
-				if v[len("version:"):] == "902" {
-					// Workaround for https://bugs.llvm.org/show_bug.cgi?id=36968, 'g' command crashes a version of debugserver on some systems (?)
-					p.gcmdok = false
-					break
-				}
-
-				// Workaround for darwin arm64. Apple's debugserver seems to have a problem
-				// with not selecting the correct thread in the 'g' command and the returned
-				// registers are empty / an E74 error is thrown. This was reported to LLVM
-				// as https://bugs.llvm.org/show_bug.cgi?id=50169
-				version, err := strconv.ParseInt(v[len("version:"):], 10, 32)
-
-				if err == nil && version >= 1200 && version <= 1205 && p.bi.Arch.Name == "arm64" {
-					p.gcmdok = false
-				}
-			}
-		}
+	if p.conn.isDebugserver {
+		// There are multiple problems with the 'g'/'G' commands on debugserver.
+		// On version 902 it used to crash the server completely (https://bugs.llvm.org/show_bug.cgi?id=36968),
+		// on arm64 it results in E74 being returned (https://bugs.llvm.org/show_bug.cgi?id=50169)
+		// and on systems where AVX-512 is used it returns the floating point
+		// registers scrambled and sometimes causes the mask registers to be
+		// zeroed out (https://github.com/go-delve/delve/pull/2498).
+		// All of these bugs stem from the fact that the main consumer of
+		// debugserver, lldb, never uses 'g' or 'G' which would make Delve the
+		// sole tester of those codepaths.
+		// Therefore we disable it here. The associated code is kept around to be
+		// used with Mozilla RR.
+		p.gcmdok = false
 	}
 
 	tgt, err := p.initialize(path, debugInfoDirs, stopReason)
@@ -365,6 +371,18 @@ func (p *gdbProcess) Connect(conn net.Conn, path string, pid int, debugInfoDirs
 	return tgt, nil
 }
 
+func (p *gdbProcess) SupportsBPF() bool {
+	return false
+}
+
+func (dbp *gdbProcess) GetBufferedTracepoints() []ebpf.RawUProbeParams {
+	return nil
+}
+
+func (dbp *gdbProcess) SetUProbe(fnName string, goidOffset int64, args []ebpf.UProbeArgMap) error {
+	panic("not implemented")
+}
+
 // unusedPort returns an unused tcp port
 // This is a hack and subject to a race condition with other running
 // programs, but most (all?) OS will cycle through all ephemeral ports
@@ -399,7 +417,7 @@ func getDebugServerAbsolutePath() string {
 
 func canUnmaskSignals(debugServerExecutable string) bool {
 	checkCanUnmaskSignalsOnce.Do(func() {
-		buf, _ := exec.Command(debugServerExecutable, "--unmask-singals").CombinedOutput()
+		buf, _ := exec.Command(debugServerExecutable, "--unmask-signals").CombinedOutput()
 		canUnmaskSignalsCached = !strings.Contains(string(buf), "unrecognized option")
 	})
 	return canUnmaskSignalsCached
@@ -436,6 +454,9 @@ func LLDBLaunch(cmd []string, wd string, flags proc.LaunchFlags, debugInfoDirs [
 	if runtime.GOOS == "windows" {
 		return nil, ErrUnsupportedOS
 	}
+	if err := macutil.CheckRosetta(); err != nil {
+		return nil, err
+	}
 
 	foreground := flags&proc.LaunchForeground != 0
 
@@ -553,6 +574,9 @@ func LLDBAttach(pid int, path string, debugInfoDirs []string) (*proc.Target, err
 	if runtime.GOOS == "windows" {
 		return nil, ErrUnsupportedOS
 	}
+	if err := macutil.CheckRosetta(); err != nil {
+		return nil, err
+	}
 
 	var (
 		isDebugserver bool
@@ -682,7 +706,7 @@ func (p *gdbProcess) initialize(path string, debugInfoDirs []string, stopReason
 			return nil, err
 		}
 	}
-	tgt, err := proc.NewTarget(p, p.currentThread, proc.NewTargetConfig{
+	tgt, err := proc.NewTarget(p, p.conn.pid, p.currentThread, proc.NewTargetConfig{
 		Path:                path,
 		DebugInfoDirs:       debugInfoDirs,
 		DisableAsyncPreempt: runtime.GOOS == "darwin",
@@ -732,6 +756,9 @@ func (p *gdbProcess) Valid() (bool, error) {
 	if p.exited {
 		return false, proc.ErrProcessExited{Pid: p.Pid()}
 	}
+	if p.almostExited && p.conn.direction == proc.Forward {
+		return false, proc.ErrProcessExited{Pid: p.Pid()}
+	}
 	return true, nil
 }
 
@@ -768,6 +795,10 @@ const (
 	childSignal      = 0x11
 	stopSignal       = 0x13
 
+	_SIGILL = 0x4
+	_SIGFPE = 0x8
+	_SIGKILL = 0x9
+
 	debugServerTargetExcBadAccess      = 0x91
 	debugServerTargetExcBadInstruction = 0x92
 	debugServerTargetExcArithmetic     = 0x93
@@ -782,6 +813,12 @@ func (p *gdbProcess) ContinueOnce() (proc.Thread, proc.StopReason, error) {
 	if p.exited {
 		return nil, proc.StopExited, proc.ErrProcessExited{Pid: p.conn.pid}
 	}
+	if p.almostExited {
+		if p.conn.direction == proc.Forward {
+			return nil, proc.StopExited, proc.ErrProcessExited{Pid: p.conn.pid}
+		}
+		p.almostExited = false
+	}
 
 	if p.conn.direction == proc.Forward {
 		// step threads stopped at any breakpoint over their breakpoint
@@ -807,15 +844,15 @@ func (p *gdbProcess) ContinueOnce() (proc.Thread, proc.StopReason, error) {
 	var atstart bool
 continueLoop:
 	for {
-		var err error
-		var sig uint8
 		tu.Reset()
-		threadID, sig, err = p.conn.resume(p.threads, &tu)
+		sp, err := p.conn.resume(p.threads, &tu)
+		threadID = sp.threadID
 		if err != nil {
 			if _, exited := err.(proc.ErrProcessExited); exited {
 				p.exited = true
+				return nil, proc.StopExited, err
 			}
-			return nil, proc.StopExited, err
+			return nil, proc.StopUnknown, err
 		}
 
 		// For stubs that support qThreadStopInfo updateThreadList will
@@ -826,11 +863,16 @@ continueLoop:
 		if trapthread != nil && !p.threadStopInfo {
 			// For stubs that do not support qThreadStopInfo we manually set the
 			// reason the thread returned by resume() stopped.
-			trapthread.sig = sig
+			trapthread.sig = sp.sig
+			trapthread.watchAddr = sp.watchAddr
 		}
 
-		var shouldStop bool
-		trapthread, atstart, shouldStop = p.handleThreadSignals(trapthread)
+		var shouldStop, shouldExitErr bool
+		trapthread, atstart, shouldStop, shouldExitErr = p.handleThreadSignals(trapthread)
+		if shouldExitErr {
+			p.almostExited = true
+			return nil, proc.StopExited, proc.ErrProcessExited{Pid: p.conn.pid}
+		}
 		if shouldStop {
 			break continueLoop
 		}
@@ -857,21 +899,7 @@ continueLoop:
 		return nil, stopReason, fmt.Errorf("could not find thread %s", threadID)
 	}
 
-	var err error
-	switch trapthread.sig {
-	case 0x91:
-		err = errors.New("bad access")
-	case 0x92:
-		err = errors.New("bad instruction")
-	case 0x93:
-		err = errors.New("arithmetic exception")
-	case 0x94:
-		err = errors.New("emulation exception")
-	case 0x95:
-		err = errors.New("software exception")
-	case 0x96:
-		err = errors.New("breakpoint exception")
-	}
+	err := machTargetExcToError(trapthread.sig)
 	if err != nil {
 		// the signals that are reported here can not be propagated back to the target process.
 		trapthread.sig = 0
@@ -894,7 +922,7 @@ func (p *gdbProcess) findThreadByStrID(threadID string) *gdbThread {
 // and returns true if we should stop execution in response to one of the
 // signals and return control to the user.
 // Adjusts trapthread to a thread that we actually want to stop at.
-func (p *gdbProcess) handleThreadSignals(trapthread *gdbThread) (trapthreadOut *gdbThread, atstart bool, shouldStop bool) {
+func (p *gdbProcess) handleThreadSignals(trapthread *gdbThread) (trapthreadOut *gdbThread, atstart, shouldStop, shouldExitErr bool) {
 	var trapthreadCandidate *gdbThread
 
 	for _, th := range p.threads {
@@ -919,6 +947,16 @@ func (p *gdbProcess) handleThreadSignals(trapthread *gdbThread) (trapthreadOut *
 		case stopSignal: // stop
 			isStopSignal = true
 
+		case _SIGKILL:
+			if p.tracedir != "" {
+				// RR will send a synthetic SIGKILL packet right before the program
+				// exits, even if the program exited normally.
+				// Treat this signal as if the process had exited because right after
+				// this it is still possible to set breakpoints and rewind the process.
+				shouldExitErr = true
+				isStopSignal = true
+			}
+
 		// The following are fake BSD-style signals sent by debugserver
 		// Unfortunately debugserver can not convert them into signals for the
 		// process so we must stop here.
@@ -963,7 +1001,7 @@ func (p *gdbProcess) handleThreadSignals(trapthread *gdbThread) (trapthreadOut *
 		shouldStop = true
 	}
 
-	return trapthread, atstart, shouldStop
+	return trapthread, atstart, shouldStop, shouldExitErr
 }
 
 // RequestManualStop will attempt to stop the process
@@ -1045,6 +1083,7 @@ func (p *gdbProcess) Restart(pos string) (proc.Thread, error) {
 	}
 
 	p.exited = false
+	p.almostExited = false
 
 	for _, th := range p.threads {
 		th.clearBreakpointState()
@@ -1059,7 +1098,7 @@ func (p *gdbProcess) Restart(pos string) (proc.Thread, error) {
 
 	// for some reason we have to send a vCont;c after a vRun to make rr behave
 	// properly, because that's what gdb does.
-	_, _, err = p.conn.resume(nil, nil)
+	_, err = p.conn.resume(nil, nil)
 	if err != nil {
 		return nil, err
 	}
@@ -1071,8 +1110,8 @@ func (p *gdbProcess) Restart(pos string) (proc.Thread, error) {
 	p.clearThreadSignals()
 	p.clearThreadRegisters()
 
-	for addr := range p.breakpoints.M {
-		p.conn.setBreakpoint(addr, p.breakpointKind)
+	for _, bp := range p.breakpoints.M {
+		p.WriteBreakpoint(bp)
 	}
 
 	return p.currentThread, p.setCurrentBreakpoints()
@@ -1182,13 +1221,46 @@ func (p *gdbProcess) ChangeDirection(dir proc.Direction) error {
 	if p.conn.direction == dir {
 		return nil
 	}
-	if p.Breakpoints().HasInternalBreakpoints() {
+	if p.Breakpoints().HasSteppingBreakpoints() {
 		return ErrDirChange
 	}
 	p.conn.direction = dir
 	return nil
 }
 
+// StartCallInjection notifies the backend that we are about to inject a function call.
+func (p *gdbProcess) StartCallInjection() (func(), error) {
+	if p.tracedir == "" {
+		return func() {}, nil
+	}
+	if p.conn.conn == nil {
+		return nil, proc.ErrProcessExited{Pid: p.conn.pid}
+	}
+	if p.conn.direction != proc.Forward {
+		return nil, ErrStartCallInjectionBackwards
+	}
+
+	// Normally it's impossible to inject function calls in a recorded target
+	// because the sequence of instructions that the target will execute is
+	// predetermined.
+	// RR however allows this in a "diversion". When a diversion is started rr
+	// takes the current state of the process and runs it forward as a normal
+	// process, not following the recording.
+	// The gdb serial protocol does not have a way to start a diversion and gdb
+	// (the main frontend of rr) does not know how to do it. Instead a
+	// diversion is started by reading siginfo, because that's the first
+	// request gdb does when starting a function call injection.
+
+	_, err := p.conn.qXfer("siginfo", "", true)
+	if err != nil {
+		return nil, err
+	}
+
+	return func() {
+		_ = p.conn.qXferWrite("siginfo", "") // rr always returns an error for qXfer:siginfo:write... even though it works
+	}, nil
+}
+
 // GetDirection returns the current direction of execution.
 func (p *gdbProcess) GetDirection() proc.Direction {
 	return p.conn.direction
@@ -1208,15 +1280,33 @@ func (p *gdbProcess) FindBreakpoint(pc uint64) (*proc.Breakpoint, bool) {
 	return nil, false
 }
 
-func (p *gdbProcess) WriteBreakpoint(bp *proc.Breakpoint) error {
-	if bp.WatchType != 0 {
-		return errors.New("hardware breakpoints not supported")
+func watchTypeToBreakpointType(wtype proc.WatchType) breakpointType {
+	switch {
+	case wtype.Read() && wtype.Write():
+		return accessWatchpoint
+	case wtype.Write():
+		return writeWatchpoint
+	case wtype.Read():
+		return readWatchpoint
+	default:
+		return swBreakpoint
 	}
-	return p.conn.setBreakpoint(bp.Addr, p.breakpointKind)
+}
+
+func (p *gdbProcess) WriteBreakpoint(bp *proc.Breakpoint) error {
+	kind := p.breakpointKind
+	if bp.WatchType != 0 {
+		kind = bp.WatchType.Size()
+	}
+	return p.conn.setBreakpoint(bp.Addr, watchTypeToBreakpointType(bp.WatchType), kind)
 }
 
 func (p *gdbProcess) EraseBreakpoint(bp *proc.Breakpoint) error {
-	return p.conn.clearBreakpoint(bp.Addr, p.breakpointKind)
+	kind := p.breakpointKind
+	if bp.WatchType != 0 {
+		kind = bp.WatchType.Size()
+	}
+	return p.conn.clearBreakpoint(bp.Addr, watchTypeToBreakpointType(bp.WatchType), kind)
 }
 
 type threadUpdater struct {
@@ -1308,7 +1398,7 @@ func (p *gdbProcess) updateThreadList(tu *threadUpdater) error {
 
 	for _, th := range p.threads {
 		if p.threadStopInfo {
-			sig, reason, err := p.conn.threadStopInfo(th.strID)
+			sp, err := p.conn.threadStopInfo(th.strID)
 			if err != nil {
 				if isProtocolErrorUnsupported(err) {
 					p.threadStopInfo = false
@@ -1316,10 +1406,12 @@ func (p *gdbProcess) updateThreadList(tu *threadUpdater) error {
 				}
 				return err
 			}
-			th.setbp = (reason == "breakpoint" || (reason == "" && sig == breakpointSignal))
-			th.sig = sig
+			th.setbp = (sp.reason == "breakpoint" || (sp.reason == "" && sp.sig == breakpointSignal) || (sp.watchAddr > 0))
+			th.sig = sp.sig
+			th.watchAddr = sp.watchAddr
 		} else {
 			th.sig = 0
+			th.watchAddr = 0
 		}
 	}
 
@@ -1436,17 +1528,17 @@ func (t *gdbThread) Common() *proc.CommonThread {
 // StepInstruction will step exactly 1 CPU instruction.
 func (t *gdbThread) StepInstruction() error {
 	pc := t.regs.PC()
-	if _, atbp := t.p.breakpoints.M[pc]; atbp {
-		err := t.p.conn.clearBreakpoint(pc, t.p.breakpointKind)
+	if bp, atbp := t.p.breakpoints.M[pc]; atbp && bp.WatchType == 0 {
+		err := t.p.conn.clearBreakpoint(pc, swBreakpoint, t.p.breakpointKind)
 		if err != nil {
 			return err
 		}
-		defer t.p.conn.setBreakpoint(pc, t.p.breakpointKind)
+		defer t.p.conn.setBreakpoint(pc, swBreakpoint, t.p.breakpointKind)
 	}
 	// Reset thread registers so the next call to
 	// Thread.Registers will not be cached.
 	t.regs.regs = nil
-	return t.p.conn.step(t.strID, &threadUpdater{p: t.p}, false)
+	return t.p.conn.step(t, &threadUpdater{p: t.p}, false)
 }
 
 // Blocked returns true if the thread is blocked in runtime or kernel code.
@@ -1541,10 +1633,14 @@ func (regs *gdbRegisters) init(regsInfo []gdbRegisterInfo, arch *proc.Arch, regn
 	}
 	regs.buf = make([]byte, regsz)
 	for _, reginfo := range regsInfo {
-		regs.regs[reginfo.Name] = gdbRegister{regnum: reginfo.Regnum, value: regs.buf[reginfo.Offset : reginfo.Offset+reginfo.Bitsize/8]}
+		regs.regs[reginfo.Name] = regs.gdbRegisterNew(&reginfo)
 	}
 }
 
+func (regs *gdbRegisters) gdbRegisterNew(reginfo *gdbRegisterInfo) gdbRegister {
+	return gdbRegister{regnum: reginfo.Regnum, value: regs.buf[reginfo.Offset : reginfo.Offset+reginfo.Bitsize/8], ignoreOnWrite: reginfo.ignoreOnWrite}
+}
+
 // reloadRegisters loads the current value of the thread's registers.
 // It will also load the address of the thread's G.
 // Loading the address of G can be done in one of two ways reloadGAlloc, if
@@ -1572,8 +1668,7 @@ func (t *gdbThread) reloadRegisters() error {
 		}
 	}
 
-	switch t.p.bi.GOOS {
-	case "linux":
+	if t.p.bi.GOOS == "linux" {
 		if reg, hasFsBase := t.regs.regs[t.p.regnames.FsBase]; hasFsBase {
 			t.regs.gaddr = 0
 			t.regs.tls = binary.LittleEndian.Uint64(reg.value)
@@ -1612,10 +1707,19 @@ func (t *gdbThread) writeSomeRegisters(regNames ...string) error {
 }
 
 func (t *gdbThread) writeRegisters() error {
-	if t.p.gcmdok {
-		return t.p.conn.writeRegisters(t.strID, t.regs.buf)
+	if t.p.gcmdok && t.p._Gcmdok {
+		err := t.p.conn.writeRegisters(t.strID, t.regs.buf)
+		if isProtocolErrorUnsupported(err) {
+			t.p._Gcmdok = false
+		} else {
+			return err
+		}
+
 	}
 	for _, r := range t.regs.regs {
+		if r.ignoreOnWrite {
+			continue
+		}
 		if err := t.p.conn.writeRegister(t.strID, r.regnum, r.value); err != nil {
 			return err
 		}
@@ -1659,13 +1763,16 @@ func (t *gdbThread) reloadGAtPC() error {
 	// around by clearing and re-setting the breakpoint in a specific sequence
 	// with the memory writes.
 	// Additionally all breakpoints in [pc, pc+len(movinstr)] need to be removed
-	for addr := range t.p.breakpoints.M {
+	for addr, bp := range t.p.breakpoints.M {
+		if bp.WatchType != 0 {
+			continue
+		}
 		if addr >= pc && addr <= pc+uint64(len(movinstr)) {
-			err := t.p.conn.clearBreakpoint(addr, t.p.breakpointKind)
+			err := t.p.conn.clearBreakpoint(addr, swBreakpoint, t.p.breakpointKind)
 			if err != nil {
 				return err
 			}
-			defer t.p.conn.setBreakpoint(addr, t.p.breakpointKind)
+			defer t.p.conn.setBreakpoint(addr, swBreakpoint, t.p.breakpointKind)
 		}
 	}
 
@@ -1693,7 +1800,7 @@ func (t *gdbThread) reloadGAtPC() error {
 		}
 	}()
 
-	err = t.p.conn.step(t.strID, nil, true)
+	err = t.p.conn.step(t, nil, true)
 	if err != nil {
 		if err == errThreadBlocked {
 			t.regs.tls = 0
@@ -1746,7 +1853,7 @@ func (t *gdbThread) reloadGAlloc() error {
 		}
 	}()
 
-	err = t.p.conn.step(t.strID, nil, true)
+	err = t.p.conn.step(t, nil, true)
 	if err != nil {
 		if err == errThreadBlocked {
 			t.regs.tls = 0
@@ -1777,6 +1884,13 @@ func (t *gdbThread) SetCurrentBreakpoint(adjustPC bool) error {
 	// adjustPC is ignored, it is the stub's responsibiility to set the PC
 	// address correctly after hitting a breakpoint.
 	t.clearBreakpointState()
+	if t.watchAddr > 0 {
+		t.CurrentBreakpoint.Breakpoint = t.p.Breakpoints().M[t.watchAddr]
+		if t.CurrentBreakpoint.Breakpoint == nil {
+			return fmt.Errorf("could not find watchpoint at address %#x", t.watchAddr)
+		}
+		return nil
+	}
 	regs, err := t.Registers()
 	if err != nil {
 		return err
@@ -1788,13 +1902,7 @@ func (t *gdbThread) SetCurrentBreakpoint(adjustPC bool) error {
 				return err
 			}
 		}
-		t.CurrentBreakpoint = bp.CheckCondition(t)
-		if t.CurrentBreakpoint.Breakpoint != nil && t.CurrentBreakpoint.Active {
-			if g, err := proc.GetG(t); err == nil {
-				t.CurrentBreakpoint.HitCount[g.ID]++
-			}
-			t.CurrentBreakpoint.TotalHitCount++
-		}
+		t.CurrentBreakpoint.Breakpoint = bp
 	}
 	return nil
 }
@@ -1861,8 +1969,11 @@ func (t *gdbThread) SetReg(regNum uint64, reg *op.DwarfRegister) error {
 	gdbreg, ok := t.regs.regs[regName]
 	if !ok && strings.HasPrefix(regName, "xmm") {
 		// XMMn and YMMn are the same amd64 register (in different sizes), if we
-		// don't find XMMn try YMMn instead.
+		// don't find XMMn try YMMn or ZMMn instead.
 		gdbreg, ok = t.regs.regs["y"+regName[1:]]
+		if !ok {
+			gdbreg, ok = t.regs.regs["z"+regName[1:]]
+		}
 	}
 	if !ok {
 		return fmt.Errorf("could not set register %s: not found", regName)
@@ -1872,7 +1983,21 @@ func (t *gdbThread) SetReg(regNum uint64, reg *op.DwarfRegister) error {
 		return fmt.Errorf("could not set register %s: wrong size, expected %d got %d", regName, len(gdbreg.value), len(reg.Bytes))
 	}
 	copy(gdbreg.value, reg.Bytes)
-	return t.p.conn.writeRegister(t.strID, gdbreg.regnum, gdbreg.value)
+	err := t.p.conn.writeRegister(t.strID, gdbreg.regnum, gdbreg.value)
+	if err != nil {
+		return err
+	}
+	if t.p.conn.workaroundReg != nil && len(gdbreg.value) > 16 {
+		// This is a workaround for a bug in debugserver where register writes (P
+		// packet) on AVX-2 and AVX-512 registers are ignored unless they are
+		// followed by a write to an AVX register.
+		// See:
+		//  Issue #2767
+		//  https://bugs.llvm.org/show_bug.cgi?id=52362
+		reg := t.regs.gdbRegisterNew(t.p.conn.workaroundReg)
+		return t.p.conn.writeRegister(t.strID, reg.regnum, reg.value)
+	}
+	return nil
 }
 
 func (regs *gdbRegisters) Slice(floatingPoint bool) ([]proc.Register, error) {
@@ -1921,8 +2046,16 @@ func (regs *gdbRegisters) Slice(floatingPoint bool) ([]proc.Register, error) {
 
 			value := regs.regs[reginfo.Name].value
 			xmmName := "x" + reginfo.Name[1:]
-			r = proc.AppendBytesRegister(r, strings.ToUpper(xmmName), value[:16])
-			r = proc.AppendBytesRegister(r, strings.ToUpper(reginfo.Name), value[16:])
+			r = proc.AppendBytesRegister(r, strings.ToUpper(xmmName), value)
+
+		case reginfo.Bitsize == 512:
+			if !strings.HasPrefix(strings.ToLower(reginfo.Name), "zmm") || !floatingPoint {
+				continue
+			}
+
+			value := regs.regs[reginfo.Name].value
+			xmmName := "x" + reginfo.Name[1:]
+			r = proc.AppendBytesRegister(r, strings.ToUpper(xmmName), value)
 		}
 	}
 	return r, nil
@@ -1939,3 +2072,21 @@ func registerName(arch *proc.Arch, regNum uint64) string {
 	regName, _, _ := arch.DwarfRegisterToString(int(regNum), nil)
 	return strings.ToLower(regName)
 }
+
+func machTargetExcToError(sig uint8) error {
+	switch sig {
+	case 0x91:
+		return errors.New("bad access")
+	case 0x92:
+		return errors.New("bad instruction")
+	case 0x93:
+		return errors.New("arithmetic exception")
+	case 0x94:
+		return errors.New("emulation exception")
+	case 0x95:
+		return errors.New("software exception")
+	case 0x96:
+		return errors.New("breakpoint exception")
+	}
+	return nil
+}
