@@ -153,15 +153,7 @@ func (t *nativeThread) SetCurrentBreakpoint(adjustPC bool) error {
 		}
 	}
 
-	if bp != nil {
-		t.CurrentBreakpoint = bp.CheckCondition(t)
-		if t.CurrentBreakpoint.Breakpoint != nil && t.CurrentBreakpoint.Active {
-			if g, err := proc.GetG(t); err == nil {
-				t.CurrentBreakpoint.HitCount[g.ID]++
-			}
-			t.CurrentBreakpoint.TotalHitCount++
-		}
-	}
+	t.CurrentBreakpoint.Breakpoint = bp
 	return nil
 }
 
