@@ -7,9 +7,12 @@ import (
 	sys "golang.org/x/sys/windows"
 
 	"github.com/go-delve/delve/pkg/proc"
+	"github.com/go-delve/delve/pkg/proc/amd64util"
 	"github.com/go-delve/delve/pkg/proc/winutil"
 )
 
+const enableHardwareBreakpoints = false // see https://github.com/go-delve/delve/issues/2768
+
 // waitStatus is a synonym for the platform-specific WaitStatus
 type waitStatus sys.WaitStatus
 
@@ -18,6 +21,7 @@ type waitStatus sys.WaitStatus
 type osSpecificDetails struct {
 	hThread            syscall.Handle
 	dbgUiRemoteBreakIn bool // whether thread is an auxiliary DbgUiRemoteBreakIn thread created by Windows
+	delayErr           error
 }
 
 func (t *nativeThread) singleStep() error {
@@ -156,14 +160,29 @@ func (t *nativeThread) restoreRegisters(savedRegs proc.Registers) error {
 	return _SetThreadContext(t.os.hThread, savedRegs.(*winutil.AMD64Registers).Context)
 }
 
-func (t *nativeThread) writeHardwareBreakpoint(addr uint64, wtype proc.WatchType, idx uint8) error {
-	return proc.ErrHWBreakUnsupported
-}
+func (t *nativeThread) withDebugRegisters(f func(*amd64util.DebugRegisters) error) error {
+	if !enableHardwareBreakpoints {
+		return errors.New("hardware breakpoints not supported")
+	}
 
-func (t *nativeThread) clearHardwareBreakpoint(addr uint64, wtype proc.WatchType, idx uint8) error {
-	return proc.ErrHWBreakUnsupported
-}
+	context := winutil.NewCONTEXT()
+	context.ContextFlags = _CONTEXT_DEBUG_REGISTERS
 
-func (t *nativeThread) findHardwareBreakpoint() (*proc.Breakpoint, error) {
-	return nil, nil
+	err := _GetThreadContext(t.os.hThread, context)
+	if err != nil {
+		return err
+	}
+
+	drs := amd64util.NewDebugRegisters(&context.Dr0, &context.Dr1, &context.Dr2, &context.Dr3, &context.Dr6, &context.Dr7)
+
+	err = f(drs)
+	if err != nil {
+		return err
+	}
+
+	if drs.Dirty {
+		return _SetThreadContext(t.os.hThread, context)
+	}
+
+	return nil
 }
