@@ -1,6 +1,9 @@
 package version
 
-import "fmt"
+import (
+	"fmt"
+	"runtime"
+)
 
 // Version represents the current version of Delve.
 type Version struct {
@@ -14,15 +17,29 @@ type Version struct {
 var (
 	// DelveVersion is the current version of Delve.
 	DelveVersion = Version{
-		Major: "1", Minor: "6", Patch: "1", Metadata: "",
-		Build: "$Id: 114218c22f3791287c4bc2f4ff35a846a1416ee9 $",
+		Major: "1", Minor: "8", Patch: "1", Metadata: "",
+		//TODO(aarzilli): before updating this to 1.8.0 re-enable staticcheck test
+		Build: "$Id: d85f1f6b736db99a1c239c34fd4a081dcff08a3c $",
 	}
 )
 
 func (v Version) String() string {
+	fixBuild(&v)
 	ver := fmt.Sprintf("Version: %s.%s.%s", v.Major, v.Minor, v.Patch)
 	if v.Metadata != "" {
 		ver += "-" + v.Metadata
 	}
 	return fmt.Sprintf("%s\nBuild: %s", ver, v.Build)
 }
+
+var buildInfo = func() string {
+	return ""
+}
+
+func BuildInfo() string {
+	return fmt.Sprintf("%s\n%s", runtime.Version(), buildInfo())
+}
+
+var fixBuild = func(v *Version) {
+	// does nothing
+}
